from ftw.pdfgenerator.interfaces import ICustomizableLayout
from ftw.pdfgenerator.interfaces import ILayoutCustomization
from ftw.pdfgenerator.layout.makolayout import MakoLayoutBase
from zope.component import queryMultiAdapter
from zope.interface import implements


class CustomizableLayout(MakoLayoutBase):
    """For documententation see ``ICustomizableLayout``.
    When subclassing, inherit and run the testcase
    ``ftw.pdfgenerator.tests.test_customizable_layout.TestCustomizableLayout``
    """

    implements(ICustomizableLayout)

    def render_layout_template(self, args):
        customization = queryMultiAdapter(
            (self.context, self.request, self), ILayoutCustomization)

        if customization is None:
            return self.render_template(self.template_name, **args)

        customization.before_render_hook()

        # we need to re-generate the packages, since the before_render_hook
        # may have changed the packages.
        args['packages'] = self.get_packages_latex()
        args = customization.get_render_arguments(args)

        if not customization.template_name:
            return self.render_template(self.template_name, **args)
        else:
            return customization.render_template(
                customization.template_name, **args)



# pylint: disable=C0321
# C0321: More than one statement on a single line

from ftw.pdfgenerator import utils
from unittest2 import TestCase
from ftw.testing import MockTestCase
from zope.interface import Interface
from ftw.pdfgenerator.utils import provide_request_layer


class TestProvideRequestLayer(MockTestCase):

    def test_add_one_interface(self):

        request = self.create_dummy()
        class ITestLayer(Interface): pass

        provide_request_layer(request, ITestLayer)

        self.assertTrue(ITestLayer.providedBy(request))

    def test_add_list_with_interfaces(self):

        request = self.create_dummy()
        class ITestLayer(Interface): pass
        class ITestLayer2(Interface): pass

        provide_request_layer(request, [ITestLayer, ITestLayer2])

        self.assertTrue(ITestLayer.providedBy(request))
        self.assertTrue(ITestLayer2.providedBy(request))

    def test_add_two_intefaces_separately(self):

        request = self.create_dummy()
        class ITestLayer(Interface): pass
        class ITestLayer2(Interface): pass

        provide_request_layer(request, ITestLayer)
        provide_request_layer(request, ITestLayer2)

        self.assertTrue(ITestLayer.providedBy(request))
        self.assertTrue(ITestLayer2.providedBy(request))

class TestBaseclasses(TestCase):

    def test_single_object_class(self):
        class A(object): pass
        self.assertEqual(utils.baseclasses(A), [A, object])

    def test_multiple_object_classes(self):
        class A(object): pass
        class B(A): pass
        class C(B): pass
        class C2(B): pass

        self.assertEqual(utils.baseclasses(C), [C, B, A, object])
        self.assertEqual(utils.baseclasses(C2), [C2, B, A, object])

    def test_multi_inherited(self):
        class A(object): pass
        class B(object): pass
        class AB(A, B): pass
        class C(object): pass
        class ABC(AB, C): pass

        self.assertEqual(utils.baseclasses(ABC),
                         [ABC, AB, A, object, B, C])

    def test_same_baseclass(self):
        class A(object): pass
        class A1(A): pass
        class A2(A): pass
        class A12(A1, A2): pass
        class B(object): pass
        class A12B(A12, B): pass

        self.assertEqual(utils.baseclasses(A12B),
                         [A12B, A12, A1, A, object, A2, B])


class TestEntities(TestCase):

    def test_decode_htmlentites(self):
        self.assertEqual(utils.decode_htmlentities('&quot;X&gt;Y&quot;'),
                         u'"X>Y"')
        self.assertEqual(utils.decode_htmlentities('m&#38;m'),
                         u'm&m')
        self.assertEqual(utils.decode_htmlentities('a&foo;b'),
                         u'a&foo;b')

    def test_encode_htmlentities(self):
        self.assertEqual(utils.encode_htmlentities('"X>Y"'),
                         '&quot;X&gt;Y&quot;')
        self.assertEqual(utils.encode_htmlentities('m&m'),
                         'm&amp;m')
        self.assertEqual(utils.encode_htmlentities('a&foo;b'),
                         'a&amp;foo;b')
        self.assertEqual(utils.encode_htmlentities(u'uml\xe4ut'),
                         u'uml&auml;ut')
        self.assertEqual(utils.encode_htmlentities('uml\xc3\xa4ut'),
                         'uml&auml;ut')

    def test_html2xmlentities(self):
        self.assertEqual(utils.html2xmlentities('m&amp;m'),
                         'm&#38;m')
        self.assertEqual(utils.html2xmlentities('a&foo;b'),
                         'a&foo;b')

    def test_xml2htmlentities(self):
        self.assertEqual(utils.xml2htmlentities('m&#38;m'),
                         'm&amp;m')
        self.assertEqual(utils.xml2htmlentities('a&#9999;b'),
                         'a&#9999;b')





# pylint: disable=W0212, W0201
# W0212: Access to a protected member of a client class
# W0201: Attribute defined outside __init__

from ftw.pdfgenerator.customization import LayoutCustomization
from ftw.pdfgenerator.interfaces import ICustomizableLayout
from ftw.pdfgenerator.interfaces import ILayoutCustomization
from ftw.pdfgenerator.templating import BaseTemplating
from ftw.pdfgenerator.testing import PDFGENERATOR_ZCML_LAYER
from ftw.pdfgenerator.tests.example import ExampleCustomizableLayout
from ftw.pdfgenerator.tests.example import ExampleCustomization
from ftw.testing import MockTestCase
from zope.component import adaptedBy
from zope.interface import Interface
from zope.interface.verify import verifyClass
import os.path


testdata_basedir = os.path.join(os.path.dirname(__file__),
                                'templates')
templates_foo = os.path.join(testdata_basedir, 'foo')
templates_bar = os.path.join(testdata_basedir, 'bar')
templates_baz = os.path.join(testdata_basedir, 'baz')


class TestLayoutCustomization(MockTestCase):

    adapter_class = LayoutCustomization

    layer = PDFGENERATOR_ZCML_LAYER

    def setUp(self, context=None, request=None, builder=None):
        super(TestLayoutCustomization, self).setUp()
        self.context = context or self.create_dummy()
        self.request = request or self.create_dummy()
        self.builder = builder or self.create_dummy()
        self.layout = ExampleCustomizableLayout(
            self.context, self.request, self.builder)

    def test_implements_interface(self):
        self.assertTrue(ILayoutCustomization.implementedBy(
                self.adapter_class))
        verifyClass(ILayoutCustomization, self.adapter_class)

    def test_adapts_correct(self):
        descriminators = adaptedBy(self.adapter_class)

        self.assertEqual(len(descriminators), 3)

        context_iface, request_iface, layout_iface = descriminators
        self.assertTrue(issubclass(context_iface, Interface))
        self.assertTrue(issubclass(request_iface, Interface))
        self.assertTrue(issubclass(layout_iface, ICustomizableLayout))

    def test_get_render_arguments(self):
        args = {}

        adapter = self.adapter_class(
            self.context, self.request, self.layout)

        self.assertTrue(isinstance(
                adapter.get_render_arguments(args), dict))

    def test_template(self):
        adapter = self.adapter_class(
            self.context, self.request, self.layout)

        if not adapter.template_name:
            return

        template = adapter.get_raw_template(adapter.template_name)

        self.assertIn('<%inherit file="original_layout"',
                      template)

    def test_get_template_directories_merges_layout_dirs(self):
        # pylint: disable=W0223
        # W0223: Method 'render_template' is abstract in
        # class 'BaseTemplating' but is not overridden
        class Layout(BaseTemplating):
            template_directories = [templates_foo]

        class Customization(self.adapter_class):
            template_directories = [templates_bar]

        context = self.create_dummy()
        request = self.create_dummy()

        custom = Customization(context, request, Layout())
        directories = custom.get_template_directories()

        self.assertIn(templates_foo, directories)
        self.assertIn(templates_bar, directories)


class TestExampleCustomization(TestLayoutCustomization):
    adapter_class = ExampleCustomization

    def test_add_template_file_from_customization_directory(self):
        layout = self.mocker.mock()
        self.expect(layout.get_template_directories()).result(
            [templates_foo])
        self.expect(layout.template_name).result('welcome.tex')

        self.expect(layout.builder.add_file('three.txt', data='bar three\n'))

        self.replay()

        adapter = ExampleCustomization(object(), object(), layout)
        adapter.add_raw_template_file('three.txt')



# pylint: disable=E0211, E0213
# E0211: Method has no argument
# E0213: Method should have "self" as first argument

from zope.interface import Interface, Attribute


DEBUG_MODE_SESSION_KEY = 'pdfgenerator-debug-mode'


class IPDFAssembler(Interface):
    """The PDF assembler assembles the PDF. It puts together all LaTeX parts,
    merges them with the layout and builds it with the builder.
    """

    def __init__(context, request):
        """The PDF assembler is a multi-adapter adapting context and request.
        """

    def build_pdf(layout=None, builder=None, request=None):
        """Builds the LaTeX and converts it to a PDF. The pdf data is
        returned as string. If `request` is passed, it will send it writes
        it to the response using the ID of the current context as filename
        base.

        Arguments:
        layout -- Use a custom layout for this build.
        request -- Write the resulting PDF to the request.
        """

    def build_latex(layout=None, builder=None, request=None):
        """Builds the LaTeX and returns it as string without converting it to
        a PDF.

        Arguments:
        layout -- Use a custom layout for this build.
        request -- Not relevant here, but the signature should match
        `build_pdf`.
        """

    def build_zip(layout=None, builder=None, request=None):
        """Builds the LaTeX and converts it to a PDF like `build_pdf` does,
        but returns a ZIP bundle of all used files (.tex-files, images,
        resources, the resulting .pdf, etc). If `request` is passed,
        it will write the file directly to the request.

        Arguments:
        layout -- Use a custom layout for this build.
        request -- Write the resulting ZIP to the request.
        """


class IConfig(Interface):
    """PDFGenerator configuration utility.
    """

    remove_build_directory = Attribute(
        'Boolean attribute. If `True`, the build directory will be removed '
        'after finishing the build.')

    def get_build_directory():
        """Returns the path to a directory, where the PDF should be built.
        This method should not return the same path twice. The directory
        should exist and be writeable.
        """


class IBuilderFactory(Interface):
    """Factory creating a new IBuilder.
    """


class IBuilder(Interface):
    """Converts LaTeX to PDF using `pdflatex`.
    """

    def add_file(filename, data):
        """Adds a file to the build directory.
        """

    def build(latex):
        """Builds and returns the PDF.
        """

    def build_zip(latex):
        """Builds the PDF and returns the a ZIP bundle, containing the build
        directory.
        """

    def cleanup():
        """Cleanup the temporary directory. This is necessary when the
        builder was requested but nothing was built.
        """


class ILaTeXLayout(Interface):
    """A LaTeX layout defines the head of the LaTeX file and puts the
    parts of the LaTeX code together. It manages the also the packages.

    ILaTeXLayout is a multi-adapter of context, request and builder.
    """

    def __init__(context, request, builder):
        """ILaTeXLayout adapts context, request and the builder.
        """

    def use_package(packagename, options=None, append_options=True,
                    insert_after=None):
        """This will add a `\usepackage{packagname}[options]'
        command to the LaTeX code, where the options part is optionally.

        Arguments:
        `packagename` -- Name of the LaTeX package.
        `options` -- LaTeX `\usepackage` options.
        `append_options` -- If `True`, `options` will be added to existing
        options, seperated by comma (defaults to `True`). Duplicate options
        will be removed.
        `insert_after` -- Inserts the package after the package with the
        name `insert_after'.
        """

    def remove_package(packagename):
        """Removes a single package. When `packagename` is '*', all packages
        are removed.

        Arguments:
        `packagename` -- Name of the LaTeX package.
        """

    def get_packages_latex():
        """Returns the LaTeX code of the `\usepackage` commands.
        """

    def use_babel(options=None, append_options=False, **kwargs):
        """Use the "babel" package. This adds a use_package with the
        language option for this context.
        By default, the language option is guessed from the context (if
        linguaplone is used) or from the preferred language of the user.
        This behavior can be changed by passing the `options` argument,
        with either the language option as string or a list of language
        options, where the first option is the primary language.
        """

    def render_latex(content_latex):
        """Renders the layout with the `content_latex` embedded.

        Arguments:
        `content_latex` -- LaTeX code of the rendered view.
        """

    def get_builder():
        """Returns the builder instance.
        """

    def get_converter():
        """Returns the current instance of the IHTML2LaTeXConverter.
        """

    def get_views_for(obj):
        """Returns a list of `ILaTeXView`s for `obj`. If no views are
        registered for the passed object it will return an empty list.
        """

    def render_latex_for(obj):
        """Renders the LaTeX views registered for `obj`.
        """


class ICustomizableLayout(ILaTeXLayout):
    """Marks a mako layout as customizable. The template of a customizable
    layout has to insert ``${logo}`` somewhere, which may contain a
    ``\includegraphics{}`` command.
    It also has to define some mako blocks which can be filled by a
    layout customization adapter:

    ``documentclass``: contains the document class definition
    ``usePackages``: contains the use_packages statements (usually
    contains ${packages})
    ``beneathPackages``: empty block beneath ``usePackages```
    ``aboveDocument``: empty block right before ``\begin{document)``
    ``documentTop``: empty block right after ``\begin{document}``
    ``documentBottom``: empty block at the end of the document,
    before ``\end{document}``
    """


class ITemplating(Interface):
    """The `ITemplating` interface is used for mixin classes enabling
    template support for any inheriting class.

    It makes it possible to define template directories on every inheriting
    class, so that the templates easily can extend or include each other.

    The `template_directories` list can be set on every inheriting class. If
    the superclass also has set it, it will also be respected when searching
    a template. When searching a template it will walk up the superclasses
    and takes the first template found. This makes it possible to override
    tempaltes defined in superclasses.
    """

    template_directories = Attribute(
        'A list of paths to template directories. The paths in the back are '
        'dominant. The paths may be absolute or relative to the path of the '
        'module where the class is defined.')

    def get_template_directories():
        """Returns a list of absolute paths to template directories.
        The paths in front are dominant: if there are multiple templates
        with the same path in different directories, the ones from the
        directories in front of the list will be taken.
        """

    def get_template(filename):
        """Returns the contents of a template with the name `filename`
        which is found in one of the directories of
        `get_template_directories` or returns `None` if no template with
        such name was found.
        """

    def render_template(filename, **kwargs):
        """Renders a template and returns the result.
        """


class ILayoutCustomization(ITemplating):
    """Adapter interface for multi adapters adapting
    ``context, request, layout``. This adapter allows us to customize
    multiple distinct layouts defined by multiple packages without
    customizing every single layout.

    The adapter implements the ``ITemplating`` interface and has to inherit
    from ``original_layout``, which is an alias for the original layout
    template name.
    """

    def before_render_hook():
        """This hook is called before rendering the template and can be used
        for defining additional ``self.use_package()`` statements or other
        customizations.
        """

    def get_render_arguments(args):
        """With this method it is possible to add additional arguments passed
        to the template. The ``args`` argument is a dict containing the
        template arguments defined by the original layout template and my be
        extended. The method should return the modified ``args`` dict.
        """


class ILaTeXView(Interface):
    """A LaTeX view is an adapter of context, request and layout. It renders
    the object in LaTeX.
    """

    def __init__(context, request, layout):
        """A LaTeX view is a multi-adapter of context, request and layout.
        """

    def render():
        """Renders the object in LaTeX and returns the LaTeX code as string.
        """

    def convert(*args, **kwargs):
        """Convert HTML to LaTeX using the IHTML2LaTeXConverter.
        See the IHTML2LaTeXConverter.convert documentation.
        """

    def convert_plain(*args, **kwargs):
        """Convert HTML to LaTeX using the IHTML2LaTeXConverter.
        See the IHTML2LaTeXConverter.convert_plain documentation.
        """


class IRecursiveLaTeXView(ILaTeXView):
    """A recursive LaTeX view walks down the tree and renders subcontent of
    a folder too.
    """


class IHTML2LaTeXConverter(Interface):
    """A `IHTML2LatexConverter` converts HTML to LaTeX. It is a multi adapter
    of (context, request, layout, builder).

    The converter runs a list of replace statements on the HTML. Each
    statement can be either a direct replacement, a regular expression or
    a subconverter.
    """

    default_patterns = Attribute('List of default patterns.')

    def __init__(context, request, layout):
        """
        """

    def get_default_subconverters():
        """Returns the default list of subconverters.
        """

    def register_patterns(patterns):
        """Registers a list of custom patterns.
        """

    def register_subconverters(subconverters):
        """Registers a list of converters.
        """

    def get_subconverter_by_pattern(pattern):
        """Returns the currently active subconverter for the passed pattern.
        """

    def convert(html, custom_patterns=None, custom_subconverters=None,
                trim=True):
        """Converts HTML to LaTeX.

        Arguments:
        html -- HTML as string.
        custom_map -- A list of custom patterns.
        custom_converters -- A list of custom subconverters, which will be
        merged into the custom_map.
        trim -- Strip the HTML before converting to LaTeX. (Default: `True`).
        """

    def quoted_umlauts(text):
        """Replaces umlauts in text with the LaTeX's quotet umlaut notation.
        E.g. "a "o "u

        Arguments:
        text -- Text to convert
        """

    def convert_plain(text, **kwargs):
        """Converts a text to HTML and then to LaTeX.

        Arguments:
        text -- The text to convert as string.
        **kwargs -- Arguments passed to `convert()`.
        """


class IHTML2LaTeXConvertRunner(Interface):
    """A kind of fork of `IHTML2LatexConverter`, which does the actual
    converting. It is passed to sub converters, so that the can update the
    current HTML / LaTeX code.
    """

    def __init__(converter, patterns, html, trim=True):
        """
        """

    def lock_chars(startPos, endPos):
        """
        Locks a specific part of the html. Other Patterns will not match this
        part of html anymore. This is generally used by SubConverters, if
        they replaced the HTML with latex.
        See replaceAndLock()
        """

    def replace(startPos, endPos, text):
        """
        Replaces a specific part in the HTML with [text] (e.g. latex code).
        The new text will be replaced by further patterns! Please use
        lockChars() or replaceAndLock() if you don't want further patterns
        to match and replace
        """

    def replace_and_lock(startPos, endPos, text):
        """
        Replaces a spefic part in the HTML with [text] and locks it with
        lockChars() after replacing.
        See replace() and lockChars()
        """

    def convert(html, custom_patterns=None, custom_subconverters=None,
                trim=True):
        """Convert a sub-part of the HTML. This initializes another
        runner and converts returns the results. This is necessary for
        converting HTML parts matched in subconverters.
        """

    def runner_convert():
        """This method does the actual converting. It should never by called
        directly, but through HTML2LatexConverter.convert()
        """

    def quoted_umlauts(text):
        """Replaces umlauts in text with the LaTeX's quotet umlaut notation.
        E.g. "a "o "u

        Arguments:
        text -- Text to convert
        """


class ISubConverter(Interface):
    """A SubConverter converts specific, advanced parts of a HTML to LaTeX.
    """

    pattern = Attribute('Regex pattern as string.')
    placeholder = Attribute(
        'The subconverter will be registered at the point of the '
        'placeholder in the patterns list.')

    def __init__(converter, match, html):
        """
        Arguments:
        converter -- Reference to the IHTML2LaTeXConverter
        match -- regex match object
        html -- full html
        """

    def __call__():
        """Initiates the converting.
        """

    def get_html():
        """Return the matched html.
        """

    def replace_and_lock(latex):
        """Sends the `latex` back to the main converter and locks it so that
        no subsequent patterns will match / replace.
        The original match area will be replaced by `latex`.
        """

    def replace(latex):
        """Sends the `latex` back to the main converter, the original match
        area will be replaced.
        Subsequent patterns may match the latex.
        """


HTML2LATEX_MODE_REPLACE = 'replace'
HTML2LATEX_MODE_REGEXP = 'regexp'
HTML2LATEX_MODE_REGEXP_FUNCTION = 'regexp function'

HTML2LATEX_CUSTOM_PATTERN_PLACEHOLDER = 'placeholder for custom patterns'
HTML2LATEX_CUSTOM_PATTERN_PLACEHOLDER_TOP = \
    'top placeholder for custom patterns'
HTML2LATEX_CUSTOM_PATTERN_PLACEHOLDER_BOTTOM = \
    'bottom placeholder for custom patterns'

HTML2LATEX_REPEAT_MODIFIER = 'repeat pattern'
HTML2LATEX_PREVENT_CHARACTER = chr(7)



from ftw.pdfgenerator import interfaces
from ftw.pdfgenerator.html2latex import wrapper
from ftw.pdfgenerator.html2latex.patterns import DEFAULT_PATTERNS
from ftw.pdfgenerator.html2latex.subconverters import footnote
from ftw.pdfgenerator.html2latex.subconverters import htmlentities
from ftw.pdfgenerator.html2latex.subconverters import hyperlink
from ftw.pdfgenerator.html2latex.subconverters import listing
from ftw.pdfgenerator.html2latex.subconverters import table
from ftw.pdfgenerator.html2latex.subconverters import textformatting
from ftw.pdfgenerator.html2latex.subconverters import url
from ftw.pdfgenerator.utils import encode_htmlentities
from ftw.pdfgenerator.utils import xml2htmlentities
from random import choice
from zope.component import adapts
from zope.interface import implements, Interface
import re
import unicodedata


PLACEHOLDERS = (
    interfaces.HTML2LATEX_CUSTOM_PATTERN_PLACEHOLDER,
    interfaces.HTML2LATEX_CUSTOM_PATTERN_PLACEHOLDER_TOP,
    interfaces.HTML2LATEX_CUSTOM_PATTERN_PLACEHOLDER_BOTTOM,
    )

DEFAULT_PLACEHOLDER = interfaces.HTML2LATEX_CUSTOM_PATTERN_PLACEHOLDER

DEFAULT_SUBCONVERTERS = (
    table.TableConverter,
    listing.ListConverter,
    htmlentities.HtmlentitiesConverter,
    hyperlink.HyperlinkConverter,
    url.URLConverter,
    textformatting.Textformatting,
    footnote.FootnoteConverter,
    )


class BasePatternAware(object):
    """Base class providing some functions for handling a local patterns
    list.
    """

    default_patterns = None

    def __init__(self, patterns):
        self.patterns = patterns

    def register_patterns(self, patterns):
        for pattern in patterns:
            self._insert_custom_pattern(pattern)

    def register_subconverters(self, subconverters):
        for converter in subconverters:
            self._register_converter(converter)

    def get_subconverter_by_pattern(self, pattern):
        for ptn in self.patterns:
            if ptn[0] == interfaces.HTML2LATEX_MODE_REGEXP_FUNCTION and \
                    ptn[1] == pattern:
                return ptn[2]

        return None

    def _insert_custom_pattern(self, pattern,
                               placeholder=DEFAULT_PLACEHOLDER,
                               replace=True):
        """Insert a custom pattern to the pattern list. If the search term
        of the pattern already exists, the existing pattern will is updated.
        """

        if isinstance(pattern[0], wrapper.CustomPatternAtPlaceholderWrapper):
            pattern = list(pattern)
            modeObject = pattern[0]
            pattern[0] = modeObject.mode
            placeholder = modeObject.placeholder
            replace = False

        found = False
        if replace:
            for i in range(0, len(self.patterns)):
                if self.patterns[i] in PLACEHOLDERS:
                    continue

                if pattern[1] == self.patterns[i][1]:
                    # overwrite existing pattern
                    self.patterns[i] = pattern
                    found = True

        if not found:
            # pattern will be inserted at the configured placeholder
            self.patterns.insert(self.patterns.index(placeholder), pattern)

    def _register_converter(self, converter_class):
        """
        Generates a pattern with a SubConverter class.
        """
        self._insert_custom_pattern(
            pattern=(interfaces.HTML2LATEX_MODE_REGEXP_FUNCTION,
                     converter_class.pattern,
                     converter_class),
            placeholder=converter_class.placeholder)


class HTML2LatexConverter(BasePatternAware):
    implements(interfaces.IHTML2LaTeXConverter)
    adapts(Interface, Interface, interfaces.ILaTeXLayout)

    default_patterns = DEFAULT_PATTERNS

    def __init__(self, context, request, layout):
        self.context = context
        self.request = request
        self.layout = layout

        super(HTML2LatexConverter, self).__init__(
            patterns=self.__class__.default_patterns[:])
        self.register_subconverters(self.get_default_subconverters())

    def get_default_subconverters(self):
        return DEFAULT_SUBCONVERTERS

    def convert(self, html, custom_patterns=None, custom_subconverters=None,
                trim=True):

        runner = HTML2LatexConvertRunner(
            converter=self,
            patterns=self.patterns[:],
            html=html,
            trim=trim)

        if custom_patterns is not None:
            runner.register_patterns(custom_patterns)

        if custom_subconverters is not None:
            runner.register_subconverters(custom_subconverters)

        return runner.runner_convert()

    def convert_plain(self, text, **kwargs):
        html = encode_htmlentities(text)
        return self.convert(html, **kwargs)

    def quoted_umlauts(self, text):
        if isinstance(text, str):
            text = text.decode('utf-8')

        text = unicodedata.normalize('NFD', text)
        text = re.sub(ur'(.)\u0308', '"\g<1>', text, re.UNICODE)
        text = unicodedata.normalize('NFC', text)

        return text.encode('utf-8')


class HTML2LatexConvertRunner(BasePatternAware):
    implements(interfaces.IHTML2LaTeXConvertRunner)

    def __init__(self, converter, patterns, html, trim=True):
        """
        Creates a instance for converting html to latex.
        Attention: this instance should only be used ONCE for converting,
        because of the lockers and html instance attributes.
        You can use convert() on this instance, it will be proxied to the
        HTML2LatexConverter instance.
        """
        BasePatternAware.__init__(self, patterns)

        if not interfaces.IHTML2LaTeXConverter.providedBy(converter):
            raise ValueError(
                'converter should by a IHTML2LaTeXConverter')
        else:
            self.converter = converter

        self.lockers = {}
        self.html = ''
        self.patterns = patterns
        self._convert_started = False

        # we use utf8
        if type(html) == unicode:
            html = html.encode('utf8')

        # we do not use xmlentities, but htmlentities
        self.html = xml2htmlentities(html)

        # trim ?
        if trim:
            self.html = self.html.strip()

    def lock_chars(self, startPos, endPos):
        """
        Locks a specific part of the html. Other Patterns will not match this
        part of html anymore. This is generally used by SubConverters, if
        they replaced the HTML with latex.
        See replaceAndLock()
        """
        # generate new id with same length
        id_chars = 'abcdefghijklmnopqrstuvwxyz0123456789'
        stop = False
        id_ = ''
        length = endPos - startPos

        # use keys with a length of at least 10 chars for making them
        # unique enought. otherwise the key generation may loop.
        if length < 10:
            length = 10

        while not stop:
            id_ = ''.join([choice(id_chars) for _i in range(length)])
            if id_ not in self.lockers.keys() and id_ not in self.html:
                stop = True

        # lock html (replace with id)
        self.lockers[id_] = self.html[startPos:endPos]
        self.html = self.html[:startPos] + id_ + self.html[endPos:]
        return id_

    def replace(self, startPos, endPos, text):
        """
        Replaces a specific part in the HTML with [text] (e.g. latex code).
        The new text will be replaced by further patterns! Please use
        lockChars() or replaceAndLock() if you don't want further patterns
        to match and replace
        """
        self.html = self.html[:startPos] + text + self.html[endPos:]
        return self.html

    def replace_and_lock(self, startPos, endPos, text):
        """
        Replaces a spefic part in the HTML with [text] and locks it with
        lockChars() after replacing.
        See replace() and lockChars()
        """
        newEndPos = startPos + len(text)
        # replace html
        self.replace(startPos, endPos, text)
        # lock chars
        return self.lock_chars(startPos, newEndPos)

    def _unlock_chars(self):
        """
        Unlocks previously locked HTML (see lockChars()). This method is
        automatically called by _convert() after converting HTML to Latex.
        """
        html_before = self.html

        for id_ in self.lockers.keys():
            value = self.lockers[id_]
            self.html = self.html.replace(id_, value)

        if html_before != self.html:
            self._unlock_chars()

    def convert(self, html, custom_patterns=None, custom_subconverters=None,
                trim=True):
        """Convert a sub-part of the HTML. This initializes another
        runner and converts returns the results. This is necessary for
        converting HTML parts matched in subconverters.
        """

        runner = HTML2LatexConvertRunner(
            converter=self.converter,
            patterns=self.patterns[:],
            html=html,
            trim=trim)

        if custom_patterns is not None:
            runner.register_patterns(custom_patterns)

        if custom_subconverters is not None:
            runner.register_subconverters(custom_subconverters)

        return runner.runner_convert()

    def runner_convert(self):
        """This method does the actual converting. It should never by called
        directly, but through HTML2LatexConverter.convert()
        """

        if self._convert_started:
            raise RuntimeError(
                'runner_convert() should not be called twice!')
        else:
            self._convert_started = True

        for pattern in self.patterns:
            if pattern in PLACEHOLDERS:
                continue

            mode = pattern[0]
            search = pattern[1]
            replace = pattern[2]
            modifiers = ()

            if len(pattern) == 4:
                modifiers = pattern[3]

            # replace
            if mode == interfaces.HTML2LATEX_MODE_REPLACE:
                self.html = self.html.replace(search, replace)

            # regexp replace
            elif mode == interfaces.HTML2LATEX_MODE_REGEXP:
                self._replace_regexp(search, replace, modifiers)

            # regexp function
            elif mode == interfaces.HTML2LATEX_MODE_REGEXP_FUNCTION:
                self._replace_regexp_function(search, replace)

        self._unlock_chars()
        return self.html

    def quoted_umlauts(self, text):
        return self.converter.quoted_umlauts(text)

    def _replace_regexp(self, search, replace, modifiers):
        xpr = re.compile(search, re.DOTALL)
        previous_html = ''
        if interfaces.HTML2LATEX_REPEAT_MODIFIER in modifiers:
            previous_html = ''

            while previous_html != self.html:
                previous_html = self.html
                self.html = xpr.sub(replace, self.html)

        else:
            self.html = xpr.sub(replace, self.html)

    def _replace_regexp_function(self, search, replace_fun):
        xpr = re.compile(search, re.DOTALL)
        skipStartPos = []
        startLimit = 0
        search = True

        while search:
            previous_html = self.html
            match = xpr.search(self.html, startLimit)

            if match and match.start() not in skipStartPos:
                skipStartPos.append(match.start())
                obj = replace_fun(self, match, self.html)
                if callable(obj):
                    obj()

            elif match and match.start() in skipStartPos:
                startLimit = match.start() + 1

            else:
                search = False

            if self.html != previous_html:
                skipStartPos = []
                startLimit = 0



from ftw.pdfgenerator import interfaces
from ftw.pdfgenerator.assembler import PDFAssembler
from ftw.pdfgenerator.testing import PDFGENERATOR_ZCML_LAYER
from ftw.testing import MockTestCase
from mocker import ARGS
from zope.component import getMultiAdapter
from zope.interface import Interface
from zope.interface.verify import verifyClass


class TestPDFAssembler(MockTestCase):

    layer = PDFGENERATOR_ZCML_LAYER

    def test_implements_interface(self):
        self.assertTrue(interfaces.IPDFAssembler.implementedBy(PDFAssembler))
        verifyClass(interfaces.IPDFAssembler, PDFAssembler)

    def test_component_registered(self):
        obj = getMultiAdapter((object(), object()), interfaces.IPDFAssembler)
        self.assertEquals(obj.__class__, PDFAssembler)

    def test_build_latex_parameters(self):
        context = object()

        layout = self.mocker.mock()
        self.expect(layout.render_latex_for(context)).result('content latex')
        self.expect(layout.render_latex('content latex')).result(
            'full latex')

        builder = self.mocker.mock()
        self.expect(builder.cleanup())

        self.replay()

        obj = getMultiAdapter((context, object()), interfaces.IPDFAssembler)
        latex = obj.build_latex(layout=layout, builder=builder)
        self.assertEqual(latex, 'full latex')

    def test_build_latex_parameters_with_fullname_parameter(self):
        context = object()

        layout = self.mocker.mock()
        self.expect(layout.render_latex_for(context)).result('content latex')
        self.expect(layout.render_latex('content latex')).result(
            'full latex')

        builder = self.mocker.mock()
        self.expect(builder.cleanup())

        self.replay()

        obj = getMultiAdapter((context, object()), interfaces.IPDFAssembler)
        latex = obj.build_latex(
            layout=layout, builder=builder, filename='test')
        self.assertEqual(latex, 'full latex')

    def test_use_filename_when_given(self):
        context = self.mocker.mock()
        layout = self.mocker.mock()
        self.expect(layout.render_latex_for(context)).result('content latex')
        self.expect(layout.render_latex('content latex')).result(
            'full latex')

        builder = self.mocker.mock()
        self.expect(builder.build('full latex')).result('the pdf')

        request = self.mocker.mock()
        response = self.mocker.mock()
        self.expect(request.RESPONSE).result(response)
        self.expect(response.setHeader(
                'Content-Type', 'application/pdf; charset=utf-8'))
        self.expect(response.setHeader(
                'Content-disposition', 'attachment; filename="my_pdf.pdf"'))
        self.expect(response.write('the pdf'))

        self.replay()

        obj = getMultiAdapter((context, request), interfaces.IPDFAssembler)
        self.assertEqual(obj.build_pdf(layout=layout, builder=builder,
                                       request=request, filename='my_pdf'),
                         request)

    def test_handles_non_ascii_chars_in_filename_correctly(self):
        filename = u'm\xe4i_pdf'
        context = self.mocker.mock()
        layout = self.mocker.mock()
        self.expect(layout.render_latex_for(context)).result('content latex')
        self.expect(layout.render_latex('content latex')).result(
            'full latex')

        builder = self.mocker.mock()
        self.expect(builder.build('full latex')).result('the pdf')

        request = self.mocker.mock()
        response = self.mocker.mock()
        self.expect(request.RESPONSE).result(response)
        self.expect(response.setHeader(
                'Content-Type', 'application/pdf; charset=utf-8'))

        self.expect(response.setHeader(
            'Content-disposition',
            'attachment; filename="{}.pdf"'.format(filename.encode('utf-8'))))
        self.expect(response.write('the pdf'))

        self.replay()

        obj = getMultiAdapter((context, request), interfaces.IPDFAssembler)
        self.assertEqual(obj.build_pdf(layout=layout, builder=builder,
                                       request=request, filename=filename),
                         request)

    def test_build_pdf_parameters(self):
        context = self.mocker.mock()
        self.expect(context.id).result('theid')

        layout = self.mocker.mock()
        self.expect(layout.render_latex_for(context)).result('content latex')
        self.expect(layout.render_latex('content latex')).result(
            'full latex')

        builder = self.mocker.mock()
        self.expect(builder.build('full latex')).result('the pdf')

        request = self.mocker.mock()
        response = self.mocker.mock()
        self.expect(request.RESPONSE).result(response)
        self.expect(response.setHeader(
                'Content-Type', 'application/pdf; charset=utf-8'))
        self.expect(response.setHeader(
                'Content-disposition', 'attachment; filename="theid.pdf"'))
        self.expect(response.write('the pdf'))

        self.replay()

        obj = getMultiAdapter((context, request), interfaces.IPDFAssembler)
        self.assertEqual(obj.build_pdf(layout=layout, builder=builder,
                                       request=request),
                         request)

    def test_build_pdf_with_no_request_returns_data(self):
        context = object()

        layout = self.mocker.mock()
        self.expect(layout.render_latex('content latex')).result(
            'full latex')
        self.expect(layout.render_latex_for(context)).result('content latex')

        builder = self.mocker.mock()
        self.expect(builder.build('full latex')).result('the pdf')

        request = object()

        self.replay()

        obj = getMultiAdapter((context, request), interfaces.IPDFAssembler)
        self.assertEqual(obj.build_pdf(layout=layout, builder=builder),
                         'the pdf')

    def test_build_zip_parameters(self):
        context = self.mocker.mock()
        self.expect(context.id, 'theid')

        layout = self.mocker.mock()
        self.expect(layout.render_latex_for(context)).result('content latex')
        self.expect(layout.render_latex('content latex')).result(
            'full latex')

        builder = self.mocker.mock()
        self.expect(builder.build_zip('full latex').read()).result('the zip')

        request = self.mocker.mock()
        response = self.mocker.mock()
        self.expect(request.RESPONSE).result(response)
        self.expect(response.setHeader(ARGS)).count(1, None)
        self.expect(response.write('the zip'))

        self.replay()

        obj = getMultiAdapter((context, request), interfaces.IPDFAssembler)
        self.assertEqual(obj.build_zip(layout=layout, builder=builder,
                                       request=request),
                         request)

    def test_build_zip_with_filename_parameter(self):
        context = self.stub()
        self.expect(context.id, 'theid')

        layout = self.mocker.mock()
        self.expect(layout.render_latex_for(context)).result('content latex')
        self.expect(layout.render_latex('content latex')).result(
            'full latex')

        builder = self.mocker.mock()
        self.expect(builder.build_zip('full latex').read()).result('the zip')

        request = self.mocker.mock()
        response = self.mocker.mock()
        self.expect(request.RESPONSE).result(response)
        self.expect(response.setHeader(ARGS)).count(1, None)
        self.expect(response.write('the zip'))

        self.replay()

        obj = getMultiAdapter((context, request), interfaces.IPDFAssembler)
        self.assertEqual(obj.build_zip(layout=layout, builder=builder,
                                       request=request, filename='my_pdf'),
                         request)

    def test_build_zip_with_no_request_returns_data(self):
        context = object()

        layout = self.mocker.mock()
        self.expect(layout.render_latex_for(context)).result('content latex')
        self.expect(layout.render_latex('content latex')).result(
            'full latex')

        builder = self.mocker.mock()
        self.expect(builder.build_zip('full latex').read()).result('the zip')

        request = object()

        self.replay()

        obj = getMultiAdapter((context, request), interfaces.IPDFAssembler)
        self.assertEqual(obj.build_zip(layout=layout, builder=builder),
                         'the zip')

    def test_get_builder(self):
        factory = self.mocker.mock()
        builder = self.mocker.mock()
        self.expect(factory()).result(builder)

        self.mock_utility(factory, interfaces.IBuilderFactory)

        self.replay()

        obj = getMultiAdapter((object(), object()), interfaces.IPDFAssembler)
        self.assertEqual(obj.get_builder(), builder)
        self.assertEqual(obj.get_builder(), obj.get_builder())

    def test_get_layout(self):
        context = self.create_dummy()
        request = self.create_dummy()
        builder = self.create_dummy()

        layout = self.mocker.mock()
        self.mock_adapter(layout, interfaces.ILaTeXLayout,
                          (Interface, Interface, Interface))
        self.expect(layout(context, request, builder)).result(layout)

        self.replay()

        obj = getMultiAdapter((context, request), interfaces.IPDFAssembler)
        obj._builder = builder
        self.assertEqual(obj.get_layout(), layout)



from ftw.pdfgenerator.html2latex.converter import HTML2LatexConverter
from unittest2 import TestCase


class SubconverterTestBase(TestCase):

    def setUp(self):
        TestCase.setUp(self)

        self.converter = HTML2LatexConverter(
            context=object(),
            request=object(),
            layout=object())

        self.convert = self.converter.convert



# pylint: disable=R0801
# R0801: Similar lines in 3 files

from ftw.pdfgenerator.interfaces import ILaTeXView, IRecursiveLaTeXView
from ftw.pdfgenerator.layout.baselayout import BaseLayout
from ftw.pdfgenerator.view import MakoLaTeXView, RecursiveLaTeXView
from ftw.testing import MockTestCase
from zope.interface import Interface
from zope.interface.verify import verifyClass
import os


class TestMakoLaTeXView(MockTestCase):

    def setUp(self):
        super(TestMakoLaTeXView, self).setUp()
        testdata_basedir = os.path.join(os.path.dirname(__file__),
                                        'templates')

        self.templates_foo = os.path.join(testdata_basedir, 'foo')

    def test_implements_interface(self):
        self.assertTrue(ILaTeXView.implementedBy(MakoLaTeXView))
        verifyClass(ILaTeXView, MakoLaTeXView)

    def test_adapted_objects_are_stored_right(self):
        context = object()
        request = object()
        layout = object()

        obj = MakoLaTeXView(context, request, layout)
        self.assertEquals(obj.context, context)
        self.assertEquals(obj.request, request)
        self.assertEquals(obj.layout, layout)

    def test_register_template_file_plain(self):
        context = object()
        request = object()
        layout = self.mocker.mock()
        self.expect(layout.get_builder().add_file('one.txt', 'foo one\n'))

        class FooView(MakoLaTeXView):
            template_directories = [self.templates_foo]

        self.replay()
        FooView(context, request, layout).register_template_file('one.txt')

    def test_register_template_file_rendererd(self):
        context = object()
        request = object()
        layout = self.mocker.mock()
        self.expect(layout.get_builder().add_file(
                'titlepage.tex',
                '\\title{Hello World}\n\n\\maketitle\n\n'))

        class FooView(MakoLaTeXView):
            template_directories = [self.templates_foo]

        self.replay()
        FooView(context, request, layout).register_template_file(
            'titlepage.tex', render=True, title='Hello World')

    def test_default_rendering_fails_if_template_name_not_defined(self):
        context = request = layout = object()

        class FooView(MakoLaTeXView):
            pass

        with self.assertRaises(ValueError) as cm:
            FooView(context, request, layout).render()

        self.assertEqual(str(cm.exception),
                         'FooView: `template_name` is not defined.')

    def test_default_rendering_with_arguments(self):
        context = request = layout = object()

        class FooView(MakoLaTeXView):
            template_directories = [self.templates_foo]
            template_name = 'titlepage.tex'

            def get_render_arguments(self):
                return {'title': 'My diary'}

        self.assertEqual(FooView(context, request, layout).render(),
                         '\\title{My diary}\n\n\\maketitle\n\n')

    def test_default_rendering_with_view_access(self):
        context = request = layout = object()

        class FooView(MakoLaTeXView):
            template_directories = [self.templates_foo]
            template_name = 'welcome.tex'

            @property
            def name(self):
                return 'Hugo Boss'

        self.assertEqual(FooView(context, request, layout).render(),
                         '{\\large Hello \\textbf{Hugo Boss}!}\n')

    def test_convert_passes_to_converter(self):
        context = request = object()
        layout = self.mocker.mock()
        converter = self.mocker.mock()

        html = 'this <b>is</b> html\n '
        latex = 'this \textbf{is} html'

        self.expect(layout.get_converter()).result(converter)
        self.expect(converter.convert(html, trim=True)).result(latex)

        self.replay()

        view = MakoLaTeXView(context, request, layout)
        self.assertEqual(view.convert(html, trim=True), latex)

    def test_convert_plain_passes_to_converter(self):
        context = request = object()
        layout = self.mocker.mock()
        converter = self.mocker.mock()

        html = 'this is <not> html\n '
        latex = 'this is <not> html'

        self.expect(layout.get_converter()).result(converter)
        self.expect(converter.convert_plain(html, trim=True)).result(latex)

        self.replay()

        view = MakoLaTeXView(context, request, layout)
        self.assertEqual(view.convert_plain(html, trim=True), latex)


class TestRecursiveLaTeXView(MockTestCase):

    def test_implements_interface(self):
        self.assertTrue(IRecursiveLaTeXView.implementedBy(RecursiveLaTeXView))
        verifyClass(IRecursiveLaTeXView, RecursiveLaTeXView)

    def test_adapted_objects_are_stored_right(self):
        context = object()
        request = object()
        layout = object()

        obj = RecursiveLaTeXView(context, request, layout)
        self.assertEquals(obj.context, context)
        self.assertEquals(obj.request, request)
        self.assertEquals(obj.layout, layout)

    def test_render_children(self):
        request = {}
        builder = object()

        context = self.mocker.mock()
        obj1 = self.mocker.mock()
        obj2 = self.mocker.mock()

        layout = BaseLayout(context, request, builder)

        self.expect(context.listFolderContents()).result([obj1, obj2])

        subview = self.mocker.mock()
        self.mock_adapter(subview, ILaTeXView,
                          (Interface, Interface, Interface))

        self.expect(subview(obj1, request, layout).render()).result(
            'object one latex')
        self.expect(subview(obj2, request, layout).render()).result(
            'object two latex')

        self.replay()

        view = RecursiveLaTeXView(context, request, layout)
        self.assertEqual(view.render_children(),
                         'object one latex\nobject two latex')

    def test_render_children__folder_contents_selected_objects(self):
        request = {'paths': [
                '/plone/parent/child1',
                '/plone/parent/child3']}
        builder = object()

        context = self.mocker.mock(count=False)
        child1 = self.mocker.mock(count=False)
        child2 = self.mocker.mock(count=False)
        child3 = self.mocker.mock(count=False)

        base_path = ['', 'plone', 'parent']
        self.expect(child1.getPhysicalPath()).result(base_path + ['child1'])
        self.expect(child2.getPhysicalPath()).result(base_path + ['child2'])
        self.expect(child3.getPhysicalPath()).result(base_path + ['child3'])

        self.expect(context.listFolderContents()).result(
            [child1, child2, child3])

        layout = BaseLayout(context, request, builder)

        subview = self.mocker.mock()
        self.mock_adapter(subview, ILaTeXView,
                          (Interface, Interface, Interface))

        self.expect(subview(child1, request, layout).render()).result(
            'child one latex')
        self.expect(subview(child3, request, layout).render()).result(
            'child three latex')

        self.replay()

        view = RecursiveLaTeXView(context, request, layout)
        self.assertEqual(
            view.render_children(),
            'child one latex\nchild three latex')

    def test_get_render_arguments_contains_latex_content(self):
        context = request = layout = object()
        view = self.mocker.patch(RecursiveLaTeXView(
                context, request, layout))

        self.expect(view.render_children()).result('children latex')

        self.replay()

        self.assertEqual(view.get_render_arguments(),
                         {'latex_content': 'children latex'})

    def test_paths_respected_when_not_on_context(self):
        request = {'paths': [
                '/plone/parent/child1']}
        builder = object()

        context = self.mocker.mock(count=False)
        child1 = self.mocker.mock(count=False)
        child1a = self.mocker.mock(count=False)

        base_path = ['', 'plone', 'parent']
        self.expect(child1.getPhysicalPath()).result(base_path + ['child1'])
        self.expect(child1a.getPhysicalPath()).result(base_path + [
                'child1', 'child1a'])

        self.expect(child1.listFolderContents()).result([child1a])

        layout = BaseLayout(context, request, builder)

        subview = self.mocker.mock()
        self.mock_adapter(subview, ILaTeXView,
                          (Interface, Interface, Interface))

        self.expect(subview(child1a, request, layout).render()).result(
            'child one A latex')

        self.replay()

        view = RecursiveLaTeXView(child1, request, layout)
        self.assertEqual(
            view.render_children(),
            'child one A latex')



# pylint: disable=W0223, C0102
# W0223: Method 'XX' is abstract in class 'YY' but is not overridden
# C0102: Black listed name "foo"



from ftw.pdfgenerator.interfaces import ITemplating
from ftw.pdfgenerator.templating import BaseTemplating
from ftw.pdfgenerator.templating import MakoTemplating
from unittest2 import TestCase
from zope.interface.verify import verifyClass
import os


class TestBaseTemplating(TestCase):

    def setUp(self):
        super(TestBaseTemplating, self).setUp()
        testdata_basedir = os.path.join(os.path.dirname(__file__),
                                        'templates')

        self.templates_foo = os.path.join(testdata_basedir, 'foo')
        self.templates_bar = os.path.join(testdata_basedir, 'bar')
        self.templates_baz = os.path.join(testdata_basedir, 'baz')

        self.templates_not_existing = os.path.join(testdata_basedir,
                                                   'not-existing')
        assert not os.path.exists(self.templates_not_existing)

    def test_implements_interface(self):
        self.assertTrue(ITemplating.implementedBy(BaseTemplating))
        verifyClass(ITemplating, BaseTemplating)

    def test_template_directories_order(self):
        class Foo(BaseTemplating):
            template_directories = [self.templates_foo,
                                    self.templates_bar]

        self.assertEqual(Foo().get_template_directories(),
                         [self.templates_bar, self.templates_foo])

    def test_template_directories_are_inherited(self):
        class Foo(BaseTemplating):
            template_directories = [self.templates_foo]

        class Bar(Foo):
            template_directories = [self.templates_bar]

        self.assertEqual(Bar().get_template_directories(),
                         [self.templates_bar,
                          self.templates_foo])

    def test_relative_directories_are_converted(self):
        class Foo(BaseTemplating):
            # his is the relative path to self.templates_foo, since it is
            # in the folder "tests".
            template_directories = ['templates/foo']

        self.assertEqual(Foo().get_template_directories(),
                         [self.templates_foo])

    def test_template_directories_is_None_does_not_fail(self):
        class Foo(BaseTemplating):
            template_directories = [self.templates_foo]

        class Bar(Foo):
            template_directories = None

        class Baz(Bar):
            template_directories = [self.templates_baz]

        self.assertEqual(Baz().get_template_directories(),
                         [self.templates_baz,
                          self.templates_foo])

    def test_template_directories_not_defined_does_not_duplicate(self):
        class Foo(BaseTemplating):
            template_directories = [self.templates_foo]

        class Bar(Foo):
            pass

        self.assertEqual(Bar().get_template_directories(),
                         [self.templates_foo])

    def test_template_directories_fails_if_definition_of_wrong_type(self):
        class Foo(BaseTemplating):
            template_directories = 1

        with self.assertRaises(ValueError) as cm:
            Foo().get_template_directories()

        self.assertEqual(
            str(cm.exception),
            'Foo.template_directories should be a list or None.')

        class Bar(BaseTemplating):
            template_directories = 1

        class Baz(Bar):
            pass

        with self.assertRaises(ValueError) as cm:
            Baz().get_template_directories()

        self.assertEqual(
            str(cm.exception),
            'Baz.template_directories should be a list or None.')

    def test_template_directories_fails_if_path_not_existing(self):
        class Foo(BaseTemplating):
            template_directories = [self.templates_not_existing]

        with self.assertRaises(OSError) as cm:
            Foo().get_template_directories()

        # OSError seems to be strange defined, so that the context manager
        # converts it to a list..
        self.assertEqual(cm.exception[1], 'No such file or directory')

    def test_get_template_returns_template(self):
        class Foo(BaseTemplating):
            template_directories = [self.templates_foo]
            # has: one.txt, two.txt, three.txt

        class Bar(Foo):
            template_directories = [self.templates_bar]
            # has: two.txt, three.txt

        class Baz(Bar):
            template_directories = [self.templates_baz]
            # has: three.txt

        obj = Baz()
        self.assertEqual(obj.get_template('one.txt'), 'foo one\n')
        self.assertEqual(obj.get_template('two.txt'), 'bar two\n')
        self.assertEqual(obj.get_template('three.txt'), 'baz three\n')

        class BarBaz(BaseTemplating):
            template_directories = [self.templates_baz,
                                    self.templates_bar]

        class FooBarBaz(BarBaz):
            template_directories = [self.templates_foo]

        obj = FooBarBaz()
        self.assertEqual(obj.get_template('one.txt'), 'foo one\n')
        self.assertEqual(obj.get_template('two.txt'), 'foo two\n')
        self.assertEqual(obj.get_template('three.txt'), 'foo three\n')

    def test_get_tempalte_returns_None_if_template_not_found(self):
        class Foo(BaseTemplating):
            template_directories = [self.templates_foo]

        self.assertEqual(Foo().get_template('not-existing.txt'), None)
        self.assertEqual(
            BaseTemplating().get_template('not-existing.txt'), None)

    def test_render_template_raises_not_implemented(self):
        # BaseTemplating does not specific a template engine - therefore
        # it is not able to render any templates. It should raise a
        # NotImplementedError.

        class Foo(BaseTemplating):
            template_directories = [self.templates_foo]

        with self.assertRaises(NotImplementedError) as cm:
            Foo().render_template('foo.txt')

        self.assertEqual(
            str(cm.exception),
            'render_template() is not implemented on BaseTemplating.')


class TestMakoTemplating(TestCase):

    def setUp(self):
        super(TestMakoTemplating, self).setUp()
        testdata_basedir = os.path.join(os.path.dirname(__file__),
                                        'templates')

        self.templates_foo = os.path.join(testdata_basedir, 'foo')
        self.templates_bar = os.path.join(testdata_basedir, 'bar')
        self.templates_baz = os.path.join(testdata_basedir, 'baz')

    def test_render_template_renders_template(self):
        class Foo(MakoTemplating):
            template_directories = [self.templates_foo]

        obj = Foo()
        self.assertEqual(
            obj.render_template('titlepage.tex', title='Wohoo'),
            '\\title{Wohoo}\n\n\\maketitle\n\n')

    def test_render_template_with_inheritance(self):
        class Foo(MakoTemplating):
            template_directories = [self.templates_foo]

        class Bar(Foo):
            template_directories = [self.templates_bar]

        obj = Bar()
        self.assertEqual(
            obj.render_template('customtitlepage.tex', title='foo',
                                custom='bar'),
            '\\title{foo}\n\n{\\large\\textbf{bar}}\n\n')

    def test_render_template_view_access(self):
        class Foo(MakoTemplating):
            template_directories = [self.templates_foo]

            def __init__(self):
                MakoTemplating.__init__(self)
                self.name = 'John'

        foo = Foo()
        self.assertEqual(
            foo.render_template('welcome.tex'),
            '{\\large Hello \\textbf{John}!}\n')

    def test_get_raw_template(self):
        class Foo(MakoTemplating):
            template_directories = [self.templates_foo]

        foo = Foo()
        self.assertEqual(
            foo.get_raw_template('welcome.tex'),
            r'{\large Hello \textbf{${view.name}}!}' + '\n')

    def test_get_raw_template_returns_None_if_template_not_found(self):
        class Foo(MakoTemplating):
            template_directories = [self.templates_foo]

        foo = Foo()
        self.assertEqual(
            foo.get_raw_template('this-file-does-not-exist'),
            None)



from ftw.pdfgenerator.customization import LayoutCustomization
from ftw.pdfgenerator.layout.customizable import CustomizableLayout
import os.path


testdata_basedir = os.path.join(os.path.dirname(__file__),
                                'templates')
templates_bar = os.path.join(testdata_basedir, 'bar')
templates_baz = os.path.join(testdata_basedir, 'baz')


class ExampleCustomizableLayout(CustomizableLayout):
    template_directories = [templates_baz]
    template_name = 'example_layout.tex'


class ExampleCustomization(LayoutCustomization):
    template_directories = [templates_bar]
    template_name = 'example_customization.tex'



from ftw.pdfgenerator.browser.standalone import BaseStandalonePDFView
from ftw.pdfgenerator.interfaces import IBuilder
from ftw.pdfgenerator.testing import PREDEFINED_BUILD_DIRECTORY_LAYER
from unittest2 import TestCase


class TestBaseStandalonePDFView(TestCase):

    layer = PREDEFINED_BUILD_DIRECTORY_LAYER

    def test_view_attributes(self):
        context = object()
        request = object()

        view = BaseStandalonePDFView(context, request)
        self.assertEquals(view.context, context)
        self.assertEquals(view.request, request)

    def test_view_is_layout(self):
        context = object()
        request = object()

        view = BaseStandalonePDFView(context, request)
        self.assertEquals(view.layout, view)

    def test_build_arguments(self):
        context = object()
        request = object()

        view = BaseStandalonePDFView(context, request)

        args = view.get_build_arguments()
        self.assertEquals(set(args.keys()),
                          set(('request', 'layout', 'builder')))

        self.assertEquals(view, args['layout'])
        self.assertEquals(view.builder, args['builder'])
        self.assertEquals(request, args['request'])

        self.assertTrue(IBuilder.providedBy(view.builder))

    def test_view_is_latex_view(self):
        context = object()
        request = object()

        view = BaseStandalonePDFView(context, request)

        self.assertEquals(view.get_views_for(context), [view])

    def test_view_does_not_render(self):
        context = object()
        request = object()

        view = BaseStandalonePDFView(context, request)

        self.assertEqual(view.render(), '')



from ftw.pdfgenerator.babel import get_preferred_babel_option_for_context
from ftw.pdfgenerator.exceptions import ConflictingUsePackageOrder
from ftw.pdfgenerator.interfaces import IBuilder
from ftw.pdfgenerator.interfaces import IHTML2LaTeXConverter
from ftw.pdfgenerator.interfaces import ILaTeXLayout
from ftw.pdfgenerator.interfaces import ILaTeXView
from zope.annotation import IAttributeAnnotatable
from zope.component import adapts, getMultiAdapter, queryMultiAdapter
from zope.interface import implements, Interface


class BaseLayout(object):
    """BaseLayout is abstract base class for all layouts. Layouts adapt
    context and request.
    """

    implements(ILaTeXLayout, IAttributeAnnotatable)
    adapts(Interface, Interface, IBuilder)

    def __init__(self, context, request, builder):
        self.context = context
        self.request = request
        self.builder = builder
        self._packages = []
        self._converter = None

    def use_package(self, packagename, options=None, append_options=True,
                    insert_after=None):
        """Documentation in ILaTeXLayout.use_package.
        """

        self._validate_package_name(packagename)

        if options in ('', None):
            options = []

        elif type(options) in (str, unicode):
            options = [options]

        if insert_after is not None:
            self._validate_package_name(insert_after)
            self._use_package_circular_order_validator(
                packagename, insert_after)

        package = self._get_package(packagename)
        if package is None:
            package = {'packagename': packagename,
                       'options': options,
                       'insert_after': insert_after}
            self._packages.append(package)

        else:
            package['insert_after'] = insert_after

            if options is not None and append_options:
                for opt in options:
                    if opt not in package['options']:
                        package['options'].append(opt)

            elif options is not None and not append_options:
                package['options'] = options

    def remove_package(self, packagename):
        """Documentation in ILaTeXLayout.remove_package.
        """

        self._validate_package_name(packagename)

        if packagename == '*':
            self._packages = []

        else:
            package = self._get_package(packagename)
            if package is not None:
                self._packages.remove(package)

    def get_packages_latex(self):
        """Documentation in ILaTeXLayout.get_packages_latex.
        """
        latex = []

        for package in self._get_ordered_packages():
            pkg_latex = r'\usepackage'

            if package['options']:
                pkg_latex += '[%s]' % ', '.join(package['options'])

            pkg_latex += '{%s}\n' % package['packagename']
            latex.append(pkg_latex)

        return ''.join(latex)

    def use_babel(self, options=None, append_options=False, **kwargs):
        """Use the "babel" package. This adds a use_package with the
        language option for this context.
        By default, the language option is guessed from the context (if
        linguaplone is used) or from the preferred language of the user.
        This behavior can be changed by passing the `options` argument,
        with either the language option as string or a list of language
        options, where the first option is the primary language.
        """

        if options is None:
            option = get_preferred_babel_option_for_context(self.context)
            if option:
                options = [option]

        elif isinstance(options, (str, unicode)):
            options = [options]

        elif hasattr(options, '__iter__'):
            options.reverse()

        kwargs['append_options'] = append_options
        kwargs['options'] = options
        self.use_package('babel', **kwargs)

    def render_latex(self, content_latex):
        """Renders the layout with the `content_latex` embedded.

        Arguments:
        `content_latex` -- LaTeX code of the rendered view.
        """
        raise NotImplementedError()

    def get_builder(self):
        """Documentation in ILaTeXLayout.get_builder
        """
        return self.builder

    def _validate_package_name(self, packagename):
        """Validates the type of a package. It should be string or unicode.
        """
        if isinstance(packagename, unicode):
            packagename = packagename.encode('utf-8')

        elif not isinstance(packagename, str):
            raise ValueError(
                'Package name should be a string, got %s (%s)' % (
                    str(packagename), type(packagename).__name__))

    def _use_package_circular_order_validator(self, packagename,
                                              dependent_packagename):
        """Test if ..
        - packagename has insert_after to itself
        - there are any insert_after loops.
        """

        if packagename == dependent_packagename:
            raise ConflictingUsePackageOrder(
                'Cannot insert "%s" after itself (insert_after="%s").' % (
                    packagename, dependent_packagename))

        path = [packagename]
        package = self._get_package(dependent_packagename)

        while package is not None:
            path.append(package['packagename'])
            if package['insert_after'] in path:
                raise ConflictingUsePackageOrder(
                    'Conflicting order: %s.' % (
                        ' after '.join(path + [package['insert_after']])))

            else:
                package = self._get_package(package['insert_after'])

        return True

    def _get_package(self, packagename):
        """Returns a package from the package configuration or None if it is
        not registered yet.
        """

        if packagename is None:
            return None

        for package in self._packages:
            if package['packagename'] == packagename:
                return package

        return None

    def _get_ordered_packages(self):
        """Returns the packages ordered by the insert_after
        definitions.
        """

        order = []

        for package in self._packages:
            if package['insert_after']:
                order.append(package['insert_after'])

        for package in self._packages:
            if package['packagename'] not in order:
                order.append(package['packagename'])

        packages = self._packages[:]
        packages.sort(lambda a, b: cmp(order.index(a['packagename']),
                                       order.index(b['packagename'])))

        return packages

    def get_converter(self):
        """Returns the current instance of the IHTML2LaTeXConverter.
        """

        if getattr(self, '_converter', None) is None:
            self._converter = getMultiAdapter(
                (self.context, self.request, self),
                IHTML2LaTeXConverter)

        return self._converter

    def get_views_for(self, obj):
        """Returns a list of `ILaTeXView`s for `obj`. If no views are
        registered for the passed object it will return an empty list.
        """
        views = []
        for name in ('pre-hook', '', 'post-hook'):
            view = queryMultiAdapter((obj, self.request, self),
                                     ILaTeXView,
                                     name=name)
            if view:
                views.append(view)

        return views

    def render_latex_for(self, obj):
        """Renders the LaTeX views registered for `obj`.
        """
        latex = []

        for view in self.get_views_for(obj):
            latex.append(view.render())

        return '\n'.join(latex)



from ftw.pdfgenerator.html2latex import subconverter
from ftw.pdfgenerator.interfaces import HTML2LATEX_PREVENT_CHARACTER
from Products.CMFCore.utils import getToolByName
from urlparse import urlparse
from urlparse import urlunparse
import os.path
import re


class URLConverter(subconverter.SubConverter):
    """Converts URLs within text (not link tags) so that they
    support hyphenation.
    """

    # http://stackoverflow.com/questions/6883049/regex-to-find-urls-in-string-in-python
    pattern = r'http[s]?://(?:[a-zA-Z]|[0-9]|[$-_@.&+]|[!*\(\),]|(?:%[0-9a-fA-F][0-9a-fA-F]))+'

    def __call__(self):
        url = self.get_html()
        scheme, netloc, path, params, query, fragment = urlparse(url)
        path = path.replace('/', '""/')
        url = urlunparse((scheme, netloc, path, params, query, fragment))
        url = url.replace('/""/', '""/', 1)
        self.replace_and_lock(url)



from ftw.pdfgenerator.html2latex.converter import HTML2LatexConverter
from ftw.pdfgenerator.html2latex.subconverters import table
from ftw.pdfgenerator.testing import ZCML_WITH_SITE_LAYER
from ftw.testing import MockTestCase
from unittest2 import TestCase


class TestTableConverter(MockTestCase):

    layer = ZCML_WITH_SITE_LAYER

    def convert(self, *args, **kwargs):
        count = 1
        if 'count' in kwargs:
            count = kwargs['count']
            del kwargs['count']

        layout = self.mocker.mock()
        self.expect(layout.use_package('multirow')).count(count)
        self.expect(layout.use_package('multicol')).count(count)
        self.expect(layout.use_package('calc')).count(0, None)

        if 'use_packages' in kwargs:
            for pkg in kwargs['use_packages']:
                self.expect(layout.use_package(pkg)).count(count)
            del kwargs['use_packages']

        self.replay()

        converter = HTML2LatexConverter(
            context=object(),
            request=object(),
            layout=layout)

        return converter.convert(*args, **kwargs)

    def test_converter_is_default(self):
        converter = HTML2LatexConverter(
            context=object(),
            request=object(),
            layout=object())

        self.assertIn(
            table.TableConverter,
            converter.get_default_subconverters())

    def test_table_converted(self):
        html = '\n'.join((
                r'<table>',
                r'    <thead>',
                r'        <tr><th>My Head</th></tr>',
                r'    </thead><tbody>',
                r'        <tr><td>My Body</td></tr>',
                r'    </tbody>',
                r'</table>'))

        latex = '\n'.join((
                r'\makeatletter\@ifundefined{tablewidth}{\newlength\tablewidth}\makeatother',
                r'\setlength\tablewidth\linewidth',
                r'\addtolength\tablewidth{-2\tabcolsep}',
                r'\renewcommand{\arraystretch}{1.4}',
                r'\begin{tabular}{l}',
                r'\multicolumn{1}{l}{\textbf{My Head}} \\',
                r'\multicolumn{1}{l}{My Body} \\',
                r'\end{tabular}\\',
                r'\vspace{4pt}',
                r''))

        self.assertMultiLineEqual(self.convert(html), latex)

    def test_table_converted_with_non_ascii(self):
        # Non-ASCII characters that *only* appear in table headings seem to
        # trip up BeautifulSoups charset sniffing.
        # This is a test case that would fail if BeautifulSoup was called
        # without an appropriate charset hint (fromEncoding='utf-8')
        html = '\n'.join((
                '<table>',
                '    <thead>',
                '        <tr><th>B\xc3\xa4rengraben</th></tr>',
                '    </thead><tbody>',
                '        <tr><td>My Body</td></tr>',
                '    </tbody>',
                '</table>'))

        latex = '\n'.join((
                '\\makeatletter\\@ifundefined{tablewidth}{\\newlength\\tablewidth}\\makeatother',
                '\\setlength\\tablewidth\\linewidth',
                '\\addtolength\\tablewidth{-2\\tabcolsep}',
                '\\renewcommand{\\arraystretch}{1.4}',
                '\\begin{tabular}{l}',
                '\\multicolumn{1}{l}{\\textbf{B\xc3\xa4rengraben}} \\\\',
                '\\multicolumn{1}{l}{My Body} \\\\',
                '\\end{tabular}\\\\',
                '\\vspace{4pt}',
                ''))

        self.assertMultiLineEqual(self.convert(html), latex)

    def test_headings(self):
        html = '\n'.join((
                r'<table>',
                r'    <thead>',
                r'        <tr>',
                r'            <td>headA</td>',
                r'            <td>headB</td>',
                r'        </tr>',
                r'    </thead>',
                r'    <tbody>',
                r'        <tr>',
                r'            <td>1A</td>',
                r'            <td>1B</td>',
                r'        </tr>',
                r'        <tr>',
                r'            <td>2A</td>',
                r'            <td>2B</td>',
                r'        </tr>',
                r'    </tbody>',
                r'</table>'))

        latex = '\n'.join((
                r'\makeatletter\@ifundefined{tablewidth}{\newlength\tablewidth}\makeatother',
                r'\setlength\tablewidth\linewidth',
                r'\addtolength\tablewidth{-4\tabcolsep}',
                r'\renewcommand{\arraystretch}{1.4}',
                r'\begin{tabular}{ll}',
                r'\multicolumn{1}{l}{\textbf{headA}} & ' + \
                    r'\multicolumn{1}{l}{\textbf{headB}} \\',
                r'\multicolumn{1}{l}{1A} & \multicolumn{1}{l}{1B} \\',
                r'\multicolumn{1}{l}{2A} & \multicolumn{1}{l}{2B} \\',
                r'\end{tabular}\\',
                r'\vspace{4pt}',
                r''))

        self.assertMultiLineEqual(self.convert(html), latex)

    def test_colgroup_with_width(self):
        html = '\n'.join((
                r'<table>',
                r'    <colgroup>',
                r'        <col width="30%" />',
                r'        <col width="70%" />',
                r'    </colgroup>',
                r'    <tbody>',
                r'        <tr>',
                r'            <td>test1</td>',
                r'            <td>test2</td>',
                r'        </tr><tr>',
                r'            <td>test3</td>',
                r'            <td>test4</td>',
                r'        </tr>',
                r'    </tbody>',
                r'</table>'))

        latex = '\n'.join((
                r'\makeatletter\@ifundefined{tablewidth}{\newlength\tablewidth}\makeatother',
                r'\setlength\tablewidth\linewidth',
                r'\addtolength\tablewidth{-4\tabcolsep}',
                r'\renewcommand{\arraystretch}{1.4}',
                r'\begin{tabular}{p{0.3\tablewidth}p{0.7\tablewidth}}',

                r'\multicolumn{1}{p{0.3\tablewidth}}{test1} & '
                r'\multicolumn{1}{p{0.7\tablewidth}}{test2} \\',

                r'\multicolumn{1}{p{0.3\tablewidth}}{test3} & '
                r'\multicolumn{1}{p{0.7\tablewidth}}{test4} \\',

                r'\end{tabular}\\',
                r'\vspace{4pt}',
                r''))

        self.assertMultiLineEqual(self.convert(html), latex)

    def test_colgroup_with_invalid_width(self):
        html = '\n'.join((
                r'<table>',
                r'    <colgroup>',
                r'        <col width="30%" />',
                r'        <col width="bad value" align="right" />',
                r'    </colgroup>',
                r'    <tbody>',
                r'        <tr>',
                r'            <td>test1</td>',
                r'            <td>test2</td>',
                r'        </tr>',
                r'    </tbody>',
                r'</table>'))

        latex = '\n'.join((
                r'\makeatletter\@ifundefined{tablewidth}{\newlength\tablewidth}\makeatother',
                r'\setlength\tablewidth\linewidth',
                r'\addtolength\tablewidth{-4\tabcolsep}',
                r'\renewcommand{\arraystretch}{1.4}',
                r'\begin{tabular}{p{0.3\tablewidth}r}',

                r'\multicolumn{1}{p{0.3\tablewidth}}{test1} & '
                r'\multicolumn{1}{r}{test2} \\',

                r'\end{tabular}\\',
                r'\vspace{4pt}',
                r''))

        self.assertMultiLineEqual(self.convert(html), latex)

    def test_col_and_cell_have_widths(self):
        # Colgroup sizes should work even if the alignment is defined
        # on the cell

        html = '\n'.join((
                r'<table>',
                r'    <colgroup>',
                r'        <col width="30%" />',
                r'        <col width="70%" />',
                r'    </colgroup>',
                r'    <tbody>',
                r'        <tr>',
                r'            <td align="right">test1</td>',
                r'            <td align="left">test2</td>',
                r'        </tr><tr>',
                r'            <td align="center">test3</td>',
                r'            <td>test4</td>',
                r'        </tr>',
                r'    </tbody>',
                r'</table>'))

        latex = '\n'.join((
                r'\makeatletter\@ifundefined{tablewidth}{\newlength\tablewidth}\makeatother',
                r'\setlength\tablewidth\linewidth',
                r'\addtolength\tablewidth{-4\tabcolsep}',
                r'\renewcommand{\arraystretch}{1.4}',
                r'\begin{tabular}{p{0.3\tablewidth}p{0.7\tablewidth}}',

                r'\multicolumn{1}{p{0.3\tablewidth}}{'
                r'\raggedleft test1} & \multicolumn{1}{'
                r'p{0.7\tablewidth}}{\raggedright test2} \\',

                r'\multicolumn{1}{p{0.3\tablewidth}}{\centering '
                r'test3} & \multicolumn{1}{p{0.7\tablewidth}}{test4} \\',

                r'\end{tabular}\\',
                r'\vspace{4pt}',
                r''))

        self.assertMultiLineEqual(self.convert(html), latex)

    def test_style_widths(self):

        html = '\n'.join((
                r'<table>',
                r'    <tbody>',
                r'        <tr>',
                r'            <td align="right" style="width:30%">test1</td>',
                r'            <td align="left" ' +
                r'style="color: red; width: 70%;">test2</td>',
                r'        </tr><tr>',
                r'            <td align="center">test3</td>',
                r'            <td>test4</td>',
                r'        </tr>',
                r'    </tbody>',
                r'</table>'))

        latex = '\n'.join((
                r'\makeatletter\@ifundefined{tablewidth}{' +
                r'\newlength\tablewidth}\makeatother',
                r'\setlength\tablewidth\linewidth',
                r'\addtolength\tablewidth{-4\tabcolsep}',
                r'\renewcommand{\arraystretch}{1.4}',
                r'\begin{tabular}{p{0.3\tablewidth}p{0.7\tablewidth}}',

                r'\multicolumn{1}{p{0.3\tablewidth}}{'
                r'\raggedleft test1} & \multicolumn{1}{'
                r'p{0.7\tablewidth}}{\raggedright test2} \\',

                r'\multicolumn{1}{p{0.3\tablewidth}}{\centering '
                r'test3} & \multicolumn{1}{p{0.7\tablewidth}}{test4} \\',

                r'\end{tabular}\\',
                r'\vspace{4pt}',
                r''))

        self.assertMultiLineEqual(self.convert(html), latex)

    def test_style_widths_in_px(self):

        html = '\n'.join((
                r'<table>',
                r'    <tbody>',
                r'        <tr>',
                r'            <td align="right" style="width:30px">test1</td>',
                r'            <td align="left" ' +
                r'style="color: red; width: 70px;">test2</td>',
                r'        </tr><tr>',
                r'            <td align="center">test3</td>',
                r'            <td>test4</td>',
                r'        </tr>',
                r'    </tbody>',
                r'</table>'))

        latex = '\n'.join((
                r'\makeatletter\@ifundefined{tablewidth}{' +
                r'\newlength\tablewidth}\makeatother',
                r'\setlength\tablewidth\linewidth',
                r'\addtolength\tablewidth{-4\tabcolsep}',
                r'\renewcommand{\arraystretch}{1.4}',
                r'\begin{tabular}{p{30.0em}p{70.0em}}',

                r'\multicolumn{1}{p{30.0em}}{'
                r'\raggedleft test1} & \multicolumn{1}{'
                r'p{70.0em}}{\raggedright test2} \\',

                r'\multicolumn{1}{p{30.0em}}{\centering '
                r'test3} & \multicolumn{1}{p{70.0em}}{test4} \\',

                r'\end{tabular}\\',
                r'\vspace{4pt}',
                r''))

        self.assertMultiLineEqual(self.convert(html), latex)


    def test_caption_is_used(self):
        # The table caption is used and will be shown in the table index

        html = '\n'.join((
                r'<table>',
                r'    <caption>Testtabelle</caption>',
                r'    <tr><td>test</td></tr>',
                r'</table>'))

        latex = '\n'.join((
                r'\makeatletter\@ifundefined{tablewidth}{' + \
                    r'\newlength\tablewidth}\makeatother',
                r'\setlength\tablewidth\linewidth',
                r'\addtolength\tablewidth{-2\tabcolsep}',
                r'\renewcommand{\arraystretch}{1.4}',

                r'\begin{center}',
                r'\addtocounter{table}{1}',
                r'\addcontentsline{lot}{table}{' + \
                    r'\protect\numberline ' + \
                    r'{\thechapter.\arabic{table}}' + \
                    r'{\ignorespaces Testtabelle}' + \
                    r'}',
                r'Table \thechapter.\arabic{table}: Testtabelle',
                r'\end{center}',
                r'\vspace{-\baselineskip}',

                r'\begin{tabular}{l}',
                r'\multicolumn{1}{l}{test} \\',
                r'\end{tabular}\\',
                r'\vspace{4pt}',
                r''))

        self.assertMultiLineEqual(self.convert(html), latex)

    def test_caption_not_listed(self):
        # It is also possible to add a caption whithout showing the table
        # in the table index by adding the css class "notListed"

        html = '\n'.join((
                r'<table class="notListed">',
                r'    <caption>NotIndexedCaption</caption>',
                r'    <tr><td>foo</td></tr>',
                r'</table>'))

        latex = '\n'.join((
                r'\makeatletter\@ifundefined{tablewidth}{' + \
                    r'\newlength\tablewidth}\makeatother',
                r'\setlength\tablewidth\linewidth',
                r'\addtolength\tablewidth{-2\tabcolsep}',
                r'\renewcommand{\arraystretch}{1.4}',

                r'\begin{center}',
                r'NotIndexedCaption',
                r'\end{center}',
                r'\vspace{-\baselineskip}',

                r'\begin{tabular}{l}',
                r'\multicolumn{1}{l}{foo} \\',
                r'\end{tabular}\\',
                r'\vspace{4pt}',
                r''))

        self.assertMultiLineEqual(self.convert(html), latex)

    def test_caption_at_bottom(self):
        # For putting the caption to the bottom of table, just at the
        # <caption>-Tag at the bottom:

        html = '\n'.join((
                r'<table>',
                r'    <tr><td>foo</td></tr>',
                r'    <caption>My Table</caption>',
                r'</table>'))

        latex = '\n'.join((
                r'\makeatletter\@ifundefined{tablewidth}{' + \
                    r'\newlength\tablewidth}\makeatother',
                r'\setlength\tablewidth\linewidth',
                r'\addtolength\tablewidth{-2\tabcolsep}',
                r'\renewcommand{\arraystretch}{1.4}',

                r'\begin{tabular}{l}',
                r'\multicolumn{1}{l}{foo} \\',
                r'\end{tabular}\\',
                r'\vspace{4pt}',

                r'\vspace{-\baselineskip}',
                r'\begin{center}',
                r'\addtocounter{table}{1}',
                r'\addcontentsline{lot}{table}{' + \
                    r'\protect\numberline ' + \
                    r'{\thechapter.\arabic{table}}' + \
                    r'{\ignorespaces My Table}' + \
                    r'}',
                r'Table \thechapter.\arabic{table}: My Table',
                r'\end{center}',

                r''))

        self.assertMultiLineEqual(self.convert(html), latex)

    def test_summary_as_caption(self):
        # Tinymce allows easily to put a <caption> tag at the top but not
        # at the bottom.
        # Therefore we use the "summary"-Attribute as bottom-caption.

        html = '\n'.join((
                r'<table summary="a quite simple table">',
                r'    <tr><td>foo</td></tr>',
                r'</table>'))

        latex = '\n'.join((
                r'\makeatletter\@ifundefined{tablewidth}{' + \
                    r'\newlength\tablewidth}\makeatother',
                r'\setlength\tablewidth\linewidth',
                r'\addtolength\tablewidth{-2\tabcolsep}',
                r'\renewcommand{\arraystretch}{1.4}',

                r'\begin{tabular}{l}',
                r'\multicolumn{1}{l}{foo} \\',
                r'\end{tabular}\\',
                r'\vspace{4pt}',

                r'\vspace{-\baselineskip}',
                r'\begin{center}',
                r'\addtocounter{table}{1}',
                r'\addcontentsline{lot}{table}{' + \
                    r'\protect\numberline ' + \
                    r'{\thechapter.\arabic{table}}' + \
                    r'{\ignorespaces a quite simple table}' + \
                    r'}',
                r'Table \thechapter.\arabic{table}: a quite simple table',
                r'\end{center}',

                r''))

        self.assertMultiLineEqual(self.convert(html), latex)

    def test_summary_as_not_listed_caption(self):
        html = '\n'.join((
                r'<table class="notListed" summary="not listed table">',
                r'    <tr><td>foo</td></tr>',
                r'</table>'))

        latex = '\n'.join((
                r'\makeatletter\@ifundefined{tablewidth}{' + \
                    r'\newlength\tablewidth}\makeatother',
                r'\setlength\tablewidth\linewidth',
                r'\addtolength\tablewidth{-2\tabcolsep}',
                r'\renewcommand{\arraystretch}{1.4}',

                r'\begin{tabular}{l}',
                r'\multicolumn{1}{l}{foo} \\',
                r'\end{tabular}\\',
                r'\vspace{4pt}',

                r'\vspace{-\baselineskip}',
                r'\begin{center}',
                r'not listed table',
                r'\end{center}',

                r''))

        self.assertMultiLineEqual(self.convert(html), latex)

    def test_multiple_tables(self):
        # Multiple tables should work as well

        html = r'<table><tr><td>test1</td></tr></table>' + \
            r'<table><tr><td>test2</td></tr></table>'

        latex = '\n'.join((
                r'\makeatletter\@ifundefined{tablewidth}{\newlength' + \
                    r'\tablewidth}\makeatother',
                r'\setlength\tablewidth\linewidth',
                r'\addtolength\tablewidth{-2\tabcolsep}',
                r'\renewcommand{\arraystretch}{1.4}',
                r'\begin{tabular}{l}',
                r'\multicolumn{1}{l}{test1} \\',
                r'\end{tabular}\\',
                r'\vspace{4pt}',
                r'\makeatletter\@ifundefined{tablewidth}{\newlength' + \
                    r'\tablewidth}\makeatother',
                r'\setlength\tablewidth\linewidth',
                r'\addtolength\tablewidth{-2\tabcolsep}',
                r'\renewcommand{\arraystretch}{1.4}',
                r'\begin{tabular}{l}',
                r'\multicolumn{1}{l}{test2} \\',
                r'\end{tabular}\\',
                r'\vspace{4pt}',
                r''))

        self.assertMultiLineEqual(self.convert(html, count=2), latex)

    def test_convert_cell_html(self):
        # Table cells can contain other HTML which will be converted as well

        html = r'<table><tr><td><b>Hello</b> <i>World</i></td></tr></table>'

        latex = '\n'.join((
                r'\makeatletter\@ifundefined{tablewidth}{\newlength\tablewidth}\makeatother',
                r'\setlength\tablewidth\linewidth',
                r'\addtolength\tablewidth{-2\tabcolsep}',
                r'\renewcommand{\arraystretch}{1.4}',
                r'\begin{tabular}{l}',
                r'\multicolumn{1}{l}{\textbf{Hello} \textit{World}} \\',
                r'\end{tabular}\\',
                r'\vspace{4pt}',
                r''))

        self.assertMultiLineEqual(self.convert(html), latex)

    def test_convert_other_html(self):
        # HTML around the Table should be converted as well

        html = '\n'.join((
                r'This <i>is</i> a <b>Table</b>:<br />',
                r'<table><tr><td>test</td></tr></table>',
                r'<b>yeah</b>'))

        latex = '\n'.join((
                r'This \textit{is} a \textbf{Table}:\\',
                r' \makeatletter\@ifundefined{tablewidth}{\newlength\tablewidth}\makeatother',
                r'\setlength\tablewidth\linewidth',
                r'\addtolength\tablewidth{-2\tabcolsep}',
                r'\renewcommand{\arraystretch}{1.4}',
                r'\begin{tabular}{l}',
                r'\multicolumn{1}{l}{test} \\',
                r'\end{tabular}\\',
                r'\vspace{4pt}',
                r' \textbf{yeah}'))

        self.assertMultiLineEqual(self.convert(html), latex)

    def test_htmlentities(self):
        # We should be able to use html entities (minidom has some
        # problems with htmlentites..)

        html = '\n'.join((
                r'<table>',
                r' <tr><td>2&gt;1</td></tr>',
                r' <tr><td>X&auml;Y</td></tr>',
                r'</table>'))

        latex = '\n'.join((
                r'\makeatletter\@ifundefined{tablewidth}{\newlength\tablewidth}\makeatother',
                r'\setlength\tablewidth\linewidth',
                r'\addtolength\tablewidth{-2\tabcolsep}',
                r'\renewcommand{\arraystretch}{1.4}',
                r'\begin{tabular}{l}',
                r'\multicolumn{1}{l}{2>1} \\',
                '\\multicolumn{1}{l}{X\xc3\xa4Y} \\\\',
                r'\end{tabular}\\',
                r'\vspace{4pt}',
                r''))

        self.assertMultiLineEqual(self.convert(html), latex)

    def test_colspan(self):
        # The td-Attribute "colspan" should work as well

        html = '\n'.join((
                r'<table>',
                r'    <colgroup>',
                r'        <col width="33%" />',
                r'        <col width="33%" />',
                r'        <col width="34%" />',
                r'    </colgroup>',
                r'    <tr>',
                r'        <td>one</td>',
                r'        <td>two</td>',
                r'        <td>three</td>',
                r'    </tr><tr>',
                r'        <td colspan="2">one and two</td>',
                r'        <td>three</td>',
                r'    </tr><tr>',
                r'        <td colspan="3">all three</td>',
                r'    </tr>',
                r'</table>'))

        latex = '\n'.join((
                r'\makeatletter\@ifundefined{tablewidth}{\newlength\tablewidth}\makeatother',
                r'\setlength\tablewidth\linewidth',
                r'\addtolength\tablewidth{-6\tabcolsep}',
                r'\renewcommand{\arraystretch}{1.4}',
                r'\begin{tabular}{p{0.33\tablewidth}p{0.33\tablewidth}'
                r'p{0.34\tablewidth}}',

                r'\multicolumn{1}{p{0.33\tablewidth}}{one} & '
                r'\multicolumn{1}{p{0.33\tablewidth}}{two} & '
                r'\multicolumn{1}{p{0.34\tablewidth}}{three} \\',

                r'\multicolumn{2}{p{0.66\tablewidth+2\tabcolsep}}{one and two} & '
                r'\multicolumn{1}{p{0.34\tablewidth}}{three} \\',

                r'\multicolumn{3}{p{1.0\tablewidth+4\tabcolsep}}{all three} \\',

                r'\end{tabular}\\',
                r'\vspace{4pt}',
                r''))

        self.maxDiff = None
        self.assertMultiLineEqual(self.convert(html), latex)

    def test_rowspan(self):
        # The td-Attribute "rowspan" should work as well

        html = '\n'.join((
                r'<table>',
                r'    <colgroup>',
                r'        <col width="50%"/>',
                r'        <col width="50%"/>',
                r'    </colgroup>',
                r'    <tr>',
                r'        <td rowspan="2">one</td>',
                r'        <td>two</td>',
                r'    </tr><tr>',
                r'        <td>three</td>',
                r'    </tr>',
                r'</table>'))

        latex = '\n'.join((
                r'\makeatletter\@ifundefined{tablewidth}{\newlength\tablewidth}\makeatother',
                r'\setlength\tablewidth\linewidth',
                r'\addtolength\tablewidth{-4\tabcolsep}',
                r'\renewcommand{\arraystretch}{1.4}',
                r'\begin{tabular}{p{0.5\tablewidth}p{0.5\tablewidth}}',

                r'\multirow{2}{0.5\tablewidth}{one} & \multicolumn{1}'
                r'{p{0.5\tablewidth}}{two} \\',

                r' & \multicolumn{1}{p{0.5\tablewidth}}{three} \\',
                r'\end{tabular}\\',
                r'\vspace{4pt}',
                r''))

        self.assertMultiLineEqual(self.convert(html), latex)

    def test_complex_rowspan_colspan(self):
        html = '\n'.join((
                r'<table>',
                r'  <colgroup>',
                r'    <col width="25%" />',
                r'    <col width="25%" />',
                r'    <col width="25%" />',
                r'    <col width="25%" />',
                r'  </colgroup>',
                r'  <tr>',
                r'    <td>A1</td>',
                r'    <td rowspan="2">B1-B2</td>',
                r'    <td>C1</td>',
                r'    <td>D1</td>',
                r'  </tr>',
                r'  <tr>',
                r'    <td>A2</td>',
                r'    <td rowspan="2">C2-C3</td>',
                r'    <td>D2</td>',
                r'  </tr>',
                r'  <tr>',
                r'    <td colspan="2">A3-B3</td>',
                r'    <td colspan="2">D3-E3</td>',
                r'  </tr>',
                r'</table>',
                ))

        latex = '\n'.join((
                r'\makeatletter\@ifundefined{tablewidth}{\newlength' + \
                    r'\tablewidth}\makeatother',
                r'\setlength\tablewidth\linewidth',
                r'\addtolength\tablewidth{-10\tabcolsep}',
                r'\renewcommand{\arraystretch}{1.4}',
                r'\begin{tabular}{p{0.25\tablewidth}p{0.25\tablewidth}' + \
                    r'p{0.25\tablewidth}p{0.25\tablewidth}l}',

                # row 1
                r'\multicolumn{1}{p{0.25\tablewidth}}{A1} & ' + \
                    r'\multirow{2}{0.25\tablewidth}{B1-B2} & ' + \
                    r'\multicolumn{1}{p{0.25\tablewidth}}{C1} & ' + \
                    r'\multicolumn{1}{p{0.25\tablewidth}}{D1} \\',

                # row 2
                r'\multicolumn{1}{p{0.25\tablewidth}}{A2} & ' + \
                    r' & '
                r'\multirow{2}{0.25\tablewidth}{C2-C3} & ' + \
                    r'\multicolumn{1}{p{0.25\tablewidth}}{D2} \\',

                # row 3
                r'\multicolumn{2}{p{0.5\tablewidth+2\tabcolsep}}{A3-B3} & '+\
                    r' & '
                r'\multicolumn{2}{l}{D3-E3} \\',

                r'\end{tabular}\\',
                r'\vspace{4pt}',
                r''))

        self.maxDiff = None
        self.assertMultiLineEqual(self.convert(html), latex)

    def test_rowspan_error_in_last_column(self):
        # Regression test that cells do not swap rows.

        html = '\n'.join((
                r'<table>',
                r'    <colspan>',
                r'        <col width="50%" />',
                r'        <col width="50%" />',
                r'    </colspan>',
                r'    <tbody>',
                r'        ',
                r'        <tr>',
                r'            <td>A</td>',
                r'            <td rowspan="2">2xB</td>',
                r'        </tr>',
                r'        <tr>',
                r'            <td>A</td>',
                r'        </tr>',
                r'        <tr>',
                r'            <td>A</td>',
                r'            <td>B</td>',
                r'        </tr>',
                r'        ',
                r'    </tbody>',
                r'</table>'))

        latex = '\n'.join((
                r'\makeatletter\@ifundefined{tablewidth}{' + \
                    r'\newlength\tablewidth}\makeatother',
                r'\setlength\tablewidth\linewidth',
                r'\addtolength\tablewidth{-4\tabcolsep}',
                r'\renewcommand{\arraystretch}{1.4}',
                r'\begin{tabular}{p{0.5\tablewidth}p{0.5\tablewidth}}',
                r'\multicolumn{1}{p{0.5\tablewidth}}{A} &' + \
                    r' \multirow{2}{0.5\tablewidth}{2xB} \\',
                r'\multicolumn{1}{p{0.5\tablewidth}}{A} &  \\',
                r'\multicolumn{1}{p{0.5\tablewidth}}{A} &' + \
                    r' \multicolumn{1}{p{0.5\tablewidth}}{B} \\',
                r'\end{tabular}\\',
                r'\vspace{4pt}',
                r''))

        self.maxDiff = None
        self.assertMultiLineEqual(self.convert(html), latex)

    def test_gridborder(self):

        html = '\n'.join((
                r'<table border="1">',
                r'    <colgroup>',
                r'        <col width="30%" />',
                r'        <col width="70%" />',
                r'    </colgroup>',
                r'    <tbody>',
                r'        <tr>',
                r'            <td>test1</td>',
                r'            <td>test2</td>',
                r'        </tr><tr>',
                r'            <td>test3</td>',
                r'            <td>test4</td>',
                r'        </tr>',
                r'    </tbody>',
                r'</table>'))

        latex = '\n'.join((
                r'\makeatletter\@ifundefined{tablewidth}{\newlength\tablewidth}\makeatother',
                r'\setlength\tablewidth\linewidth',
                r'\addtolength\tablewidth{-4\tabcolsep}',
                r'\renewcommand{\arraystretch}{1.4}',
                r'\begin{tabular}{|p{0.3\tablewidth}|p{0.7\tablewidth}|}',
                r'\hline',

                r'\multicolumn{1}{|p{0.3\tablewidth}}{test1} & '
                r'\multicolumn{1}{|p{0.7\tablewidth}|}{test2} \\',

                r'\hline',

                r'\multicolumn{1}{|p{0.3\tablewidth}}{test3} & '
                r'\multicolumn{1}{|p{0.7\tablewidth}|}{test4} \\',

                r'\hline',
                r'\end{tabular}\\',
                r'\vspace{4pt}',
                r''))

        self.assertMultiLineEqual(self.convert(html), latex)

    def test_newlines(self):
        # In LaTeX, there should not be a "\\" or "\n" in a table cell, but
        # a "\newline". The newline only is taken in account if the cell or
        # column is defined with a fixed width ("p{width}").

        html = '\n'.join((
                r'<table>',
                r'  <tr>',
                r'    <td width="50%">hello<br />world</td>',
                r'    <td width="50%"><p>foo</p><p>bar</p></td>',
                r'  </tr>',
                r'</table>',
                ))

        latex = '\n'.join((
                r'\makeatletter\@ifundefined{tablewidth}{\newlength' + \
                    r'\tablewidth}\makeatother',
                r'\setlength\tablewidth\linewidth',
                r'\addtolength\tablewidth{-4\tabcolsep}',
                r'\renewcommand{\arraystretch}{1.4}',
                r'\begin{tabular}{p{0.5\tablewidth}p{0.5\tablewidth}}',

                r'\multicolumn{1}{p{0.5\tablewidth}}{' + \
                    r'hello\newline world} & ' + \
                    r'\multicolumn{1}{p{0.5\tablewidth}}{' + \
                    r'foo\newline \newline bar} \\',

                r'\end{tabular}\\',
                r'\vspace{4pt}',
                r''
                ))

        self.assertMultiLineEqual(self.convert(html), latex)

    def test_tabular_environment_depends_on_table_size(self):
        # For big tables we use the "tabular" environment, allowing a table
        # to be splitted up over multiple pages, but for short tables we use
        # the longtable environment.

        tabular_html = '\n'.join((
                r'<table>',
                r'  <tr>',
                r'    <td width="50%">foo</td>',
                r'    <td width="50%">bar</td>',
                r'  </tr>',
                r'</table>',
                ))

        tabular_latex = '\n'.join((
                r'\makeatletter\@ifundefined{tablewidth}{\newlength\tablewidth}\makeatother',
                r'\setlength\tablewidth\linewidth',
                r'\addtolength\tablewidth{-4\tabcolsep}',
                r'\renewcommand{\arraystretch}{1.4}',
                r'\begin{tabular}{p{0.5\tablewidth}p{0.5\tablewidth}}',

                r'\multicolumn{1}{p{0.5\tablewidth}}{' + \
                    r'foo} & ' + \
                    r'\multicolumn{1}{p{0.5\tablewidth}}{' + \
                    r'bar} \\',

                r'\end{tabular}\\',
                r'\vspace{4pt}',
                r''
                ))

        self.assertMultiLineEqual(self.convert(tabular_html), tabular_latex)

    def test_longtable_environment_depends_on_table_size(self):
        # For big tables we use the "tabular" environment, allowing a table
        # to be splitted up over multiple pages, but for short tables we use
        # the longtable environment.

        longtable_html = '\n'.join((
                r'<table>',
                40 * r'<tr><td>foo</td></tr>',
                r'</table>',
                ))

        longtable_latex = '\n'.join((
                r'\makeatletter\@ifundefined{tablewidth}{\newlength\tablewidth}\makeatother',
                r'\setlength\tablewidth\linewidth',
                r'\addtolength\tablewidth{-2\tabcolsep}',
                r'\renewcommand{\arraystretch}{1.4}',
                r'\begin{longtable}{l}',
                (40 * '\\multicolumn{1}{l}{foo} \\\\\n').strip(),
                r'\end{longtable}\\',
                r'\vspace{4pt}',
                r''
                ))

        self.assertMultiLineEqual(self.convert(longtable_html,
                                               use_packages=['longtable']),
                                  longtable_latex)

    def test_longtable_environment_enforcable_by_cssclass(self):
        longtable_html = '\n'.join((
                r'<table class="page-break">',
                r'  <tr>',
                r'    <td width="50%">foo</td>',
                r'    <td width="50%">bar</td>',
                r'  </tr>',
                r'</table>',
                ))

        longtable_latex = '\n'.join((
                r'\makeatletter\@ifundefined{tablewidth}{\newlength\tablewidth}\makeatother',
                r'\setlength\tablewidth\linewidth',
                r'\addtolength\tablewidth{-4\tabcolsep}',
                r'\renewcommand{\arraystretch}{1.4}',
                r'\begin{longtable}{p{0.5\tablewidth}p{0.5\tablewidth}}',

                r'\multicolumn{1}{p{0.5\tablewidth}}{' + \
                    r'foo} & ' + \
                    r'\multicolumn{1}{p{0.5\tablewidth}}{' + \
                    r'bar} \\',

                r'\end{longtable}\\',
                r'\vspace{4pt}',
                r''
                ))

        self.assertMultiLineEqual(self.convert(longtable_html,
                                               use_packages=['longtable']),
                                  longtable_latex)

    def test_tabular_environment_enforcable_by_cssclass(self):
        tabular_html = '\n'.join((
                r'<table class="no-page-break">',
                20 * r'<tr><td>foo</td></tr>',
                r'</table>',
                ))

        tabular_latex = '\n'.join((
                r'\makeatletter\@ifundefined{tablewidth}{\newlength\tablewidth}\makeatother',
                r'\setlength\tablewidth\linewidth',
                r'\addtolength\tablewidth{-2\tabcolsep}',
                r'\renewcommand{\arraystretch}{1.4}',
                r'\begin{tabular}{l}',
                (20 * '\\multicolumn{1}{l}{foo} \\\\\n').strip(),
                r'\end{tabular}\\',
                r'\vspace{4pt}',
                r''
                ))

        self.assertMultiLineEqual(self.convert(tabular_html), tabular_latex)

    def test_longtable_uses_endhead_for_ths(self):
        html = '\n'.join((
                r'<table>',
                r'<tr><th>heading</th></tr>',
                20 * r'<tr><td>content</td></tr>',
                r'</table>'))

        latex = self.convert(html, use_packages=['longtable'])

        self.assertIn(r'\endhead', latex)

    def test_longtable_uses_endhead_for_thead_cells(self):
        html = '\n'.join((
                r'<table>',
                r' <thead>'
                r'  <tr><td>heading1</td></tr>',
                r'  <tr><td>heading2</td></tr>',
                r' </thead>',
                r' <tbody>',
                20 * r'<tr><td>content</td></tr>',
                r' </tbody>',
                r'</table>'))

        latex = self.convert(html, use_packages=['longtable'])

        self.assertIn(r'\endhead', latex)

    def test_tabular_uses_no_endhead(self):
        html = '\n'.join((
                r'<table class="no-page-break">',
                r' <thead>',
                r'  <tr><th>heading</th></tr>',
                r' </thead>',
                r' <tbody>',
                r'  <tr><td>content</td></tr>',
                r' </tbody>',
                r'</table>'))

        latex = self.convert(html)

        self.assertNotIn(r'\endhead', latex)

    def test_simple_grid_css_class(self):
        html = '\n'.join((
                r'<table class="no-page-break border-grid">',
                r' <thead>',
                r'  <tr>',
                r'   <th>heading A</th>',
                r'   <th>heading B</th>',
                r'  </tr>',
                r' </thead>',
                r' <tbody>',
                r'  <tr>',
                r'   <td>content 1A</td>',
                r'   <td>content 1B</td>',
                r'  </tr>',
                r'  <tr>',
                r'   <td>content 2A</td>',
                r'   <td>content 2B</td>',
                r'  </tr>',
                r' </tbody>',
                r'</table>'))

        latex = '\n'.join((
                r'\makeatletter\@ifundefined{tablewidth}{\newlength\tablewidth}\makeatother',
                r'\setlength\tablewidth\linewidth',
                r'\addtolength\tablewidth{-4\tabcolsep}',
                r'\renewcommand{\arraystretch}{1.4}',
                r'\begin{tabular}{|l|l|}',
                r'\hline',

                r'\multicolumn{1}{|l}{\textbf{heading A}} & ' + \
                    r'\multicolumn{1}{|l|}{\textbf{heading B}} \\',
                r'\hline',

                r'\multicolumn{1}{|l}{content 1A} & ' + \
                    r'\multicolumn{1}{|l|}{content 1B} \\',
                r'\hline',

                r'\multicolumn{1}{|l}{content 2A} & ' + \
                    r'\multicolumn{1}{|l|}{content 2B} \\',
                r'\hline',

                r'\end{tabular}\\',
                r'\vspace{4pt}',
                r''
                ))

        self.assertMultiLineEqual(self.convert(html), latex)

    def test_complex_grid_css_class(self):
        html = '\n'.join((
                r'<table class="no-page-break grid">',
                r' <colgroup>',
                r'  <col width="50%" />'
                r'  <col width="25%" />'
                r'  <col width="25%" />'
                r' </colgroup>',
                r' <thead>',
                r'  <tr>',
                r'   <th colspan="3">heading</th>',
                r'  </tr>',
                r' </thead>',
                r' <tbody>',
                r'  <tr>',
                r'   <td>content 1A</td>',
                r'   <td rowspan="2">content 1/2 B</td>',
                r'   <td>content 1C</td>',
                r'  </tr>',
                r'  <tr>',
                r'   <td>content 2A</td>',
                r'   <td>content 2C</td>',
                r'  </tr>',
                r' </tbody>',
                r'</table>'))

        latex = '\n'.join((
                r'\makeatletter\@ifundefined{tablewidth}{\newlength\tablewidth}\makeatother',
                r'\setlength\tablewidth\linewidth',
                r'\addtolength\tablewidth{-6\tabcolsep}',
                r'\renewcommand{\arraystretch}{1.4}',
                r'\begin{tabular}{|p{0.5\tablewidth}|p{0.25\tablewidth}|' + \
                    r'p{0.25\tablewidth}|}',
                r'\hline',

                r'\multicolumn{3}{|p{1.0\tablewidth+4\tabcolsep}|}{' + \
                    r'\textbf{heading}} \\',
                r'\hline',

                r'\multicolumn{1}{|p{0.5\tablewidth}|}{content 1A} & ' + \
                    r'\multirow{2}{0.25\tablewidth}{content 1/2 B} & ' + \
                    r'\multicolumn{1}{|p{0.25\tablewidth}|}{content 1C} \\',
                r'\cline{1-1}\cline{3-3}',

                r'\multicolumn{1}{|p{0.5\tablewidth}|}{content 2A} & ' + \
                    r' & '
                r'\multicolumn{1}{|p{0.25\tablewidth}|}{content 2C} \\',
                r'\hline',

                r'\end{tabular}\\',
                r'\vspace{4pt}',
                r''
                ))

        self.maxDiff = None
        self.assertMultiLineEqual(self.convert(html), latex)

    def test_border_rowspan_collapse(self):
        html = '\n'.join((
                r'<table class="no-page-break">',
                r' <colgroup>',
                r'  <col width="25%" />'
                r'  <col width="25%" />'
                r'  <col width="25%" />'
                r'  <col width="25%" />'
                r' </colgroup>',
                r' <tbody>',
                r'  <tr>',
                r'   <td rowspan="2">1A</td>',
                r'   <td>content 1B</td>',
                r'   <td rowspan="2" class="border-right border-left">' + \
                    r'content 1/2 C</td>',
                r'   <td>content 1D</td>',
                r'  </tr>',
                r'  <tr>',
                r'   <td>content 2B</td>',
                r'   <td>content 2D</td>',
                r'  </tr>',
                r' </tbody>',
                r'</table>'))

        latex = '\n'.join((
                r'\makeatletter\@ifundefined{tablewidth}{\newlength\tablewidth}\makeatother',
                r'\setlength\tablewidth\linewidth',
                r'\addtolength\tablewidth{-8\tabcolsep}',
                r'\renewcommand{\arraystretch}{1.4}',
                r'\begin{tabular}{p{0.25\tablewidth}p{0.25\tablewidth}' + \
                    r'p{0.25\tablewidth}p{0.25\tablewidth}}',

                r'\multirow{2}{0.25\tablewidth}{1A} & ' + \
                    r'\multicolumn{1}{p{0.25\tablewidth}|}{content 1B} & ' + \
                    r'\multirow{2}{0.25\tablewidth}{content 1/2 C} & ' + \
                    r'\multicolumn{1}{|p{0.25\tablewidth}}{content 1D} \\',

                r' & ' + \
                    r'\multicolumn{1}{p{0.25\tablewidth}|}{content 2B} & ' + \
                    r' & '
                r'\multicolumn{1}{|p{0.25\tablewidth}}{content 2D} \\',

                r'\end{tabular}\\',
                r'\vspace{4pt}',
                r''
                ))

        self.assertMultiLineEqual(self.convert(html), latex)

    def test_listing_css_class(self):
        html = '\n'.join((
                r'<table class="no-page-break listing">',
                r' <colgroup>',
                r'  <col width="50%" />'
                r'  <col width="25%" />'
                r'  <col width="25%" />'
                r' </colgroup>',
                r' <thead>',
                r'  <tr>',
                r'   <th colspan="3">heading</th>',
                r'  </tr>',
                r' </thead>',
                r' <tbody>',
                r'  <tr>',
                r'   <td>content 1A</td>',
                r'   <td rowspan="2">content 1/2 B</td>',
                r'   <td>content 1C</td>',
                r'  </tr>',
                r'  <tr>',
                r'   <td>content 2A</td>',
                r'   <td>content 2C</td>',
                r'  </tr>',
                r' </tbody>',
                r'</table>'))

        latex = '\n'.join((
                r'\makeatletter\@ifundefined{tablewidth}{\newlength\tablewidth}\makeatother',
                r'\setlength\tablewidth\linewidth',
                r'\addtolength\tablewidth{-6\tabcolsep}',
                r'\renewcommand{\arraystretch}{1.4}',
                r'\begin{tabular}{p{0.5\tablewidth}p{0.25\tablewidth}' + \
                    r'p{0.25\tablewidth}}',
                r'\hline',

                r'\multicolumn{3}{p{1.0\tablewidth+4\tabcolsep}}{' + \
                    r'\textbf{heading}} \\',
                r'\hline',

                r'\multicolumn{1}{p{0.5\tablewidth}}{content 1A} & ' + \
                    r'\multirow{2}{0.25\tablewidth}{content 1/2 B} & ' + \
                    r'\multicolumn{1}{p{0.25\tablewidth}}{content 1C} \\',
                r'\cline{1-1}\cline{3-3}',

                r'\multicolumn{1}{p{0.5\tablewidth}}{content 2A} & ' + \
                    r' & '
                r'\multicolumn{1}{p{0.25\tablewidth}}{content 2C} \\',
                r'\hline',

                r'\end{tabular}\\',
                r'\vspace{4pt}',
                r''
                ))

        self.assertMultiLineEqual(self.convert(html), latex)

    def test_vertical_css_class_makes_first_column_bold(self):
        html = '\n'.join((
                r'<table class="no-page-break vertical">',
                r' <colgroup>',
                r'  <col width="50%" />'
                r'  <col width="25%" />'
                r'  <col width="25%" />'
                r' </colgroup>',
                r' <thead>',
                r'  <tr>',
                r'   <th colspan="3">heading</th>',
                r'  </tr>',
                r' </thead>',
                r' <tbody>',
                r'  <tr>',
                r'   <td>content 1A</td>',
                r'   <td rowspan="2">content 1/2 B</td>',
                r'   <td>content 1C</td>',
                r'  </tr>',
                r'  <tr>',
                r'   <td>content 2A</td>',
                r'   <td>content 2C</td>',
                r'  </tr>',
                r' </tbody>',
                r'</table>'))

        latex = '\n'.join((
                r'\makeatletter\@ifundefined{tablewidth}{\newlength\tablewidth}\makeatother',
                r'\setlength\tablewidth\linewidth',
                r'\addtolength\tablewidth{-6\tabcolsep}',
                r'\renewcommand{\arraystretch}{1.4}',
                r'\begin{tabular}{p{0.5\tablewidth}p{0.25\tablewidth}' + \
                    r'p{0.25\tablewidth}}',

                r'\multicolumn{3}{p{1.0\tablewidth+4\tabcolsep}}{' + \
                    r'\textbf{heading}} \\',

                r'\multicolumn{1}{p{0.5\tablewidth}}' + \
                    r'{\textbf{content 1A}} & ' + \
                    r'\multirow{2}{0.25\tablewidth}{content 1/2 B} & ' + \
                    r'\multicolumn{1}{p{0.25\tablewidth}}{content 1C} \\',

                r'\multicolumn{1}{p{0.5\tablewidth}}' + \
                    r'{\textbf{content 2A}} & ' + \
                    r' & ' + \
                    r'\multicolumn{1}{p{0.25\tablewidth}}{content 2C} \\',

                r'\end{tabular}\\',
                r'\vspace{4pt}',
                r''
                ))

        self.maxDiff = None
        self.assertMultiLineEqual(self.convert(html), latex)

    def test_colspan_without_width(self):
        html = '\n'.join((
                r'<table>',
                r'<tr>',
                r'<td>foo</td>',
                r'<td>bar</td>',
                r'</tr><tr>',
                r'<td colspan="2">baz</td>',
                r'</tr>',
                r'</table>'))

        latex = '\n'.join((
                r'\makeatletter\@ifundefined{tablewidth}{' + \
                    r'\newlength\tablewidth}\makeatother',
                r'\setlength\tablewidth\linewidth',
                r'\addtolength\tablewidth{-4\tabcolsep}',
                r'\renewcommand{\arraystretch}{1.4}',
                r'\begin{tabular}{ll}',

                r'\multicolumn{1}{l}{foo} & ' + \
                    r'\multicolumn{1}{l}{bar} \\',
                r'\multicolumn{2}{l}{baz} \\',

                r'\end{tabular}\\',
                r'\vspace{4pt}',
                r''))

        self.assertMultiLineEqual(self.convert(html), latex)

    def test_border_classes(self):
        html = '\n'.join((
                r'<table class="no-page-break">',
                r' <tbody>',
                r'  <tr>',
                r'   <td class="border-right">A1</td>',
                r'   <td class="border-top">B1</td>',
                r'   <td class="border-left border-bottom">C1</td>',
                r'  </tr>',
                r'  <tr>',
                r'   <td class="border-right">A2</td>',
                r'   <td class="border-top border-bottom">B2</td>',
                r'   <td class="border-left border-bottom">C2</td>',
                r'  </tr>',
                r' </tbody>',
                r'</table>'
                ))


        latex = '\n'.join((
                r'\makeatletter\@ifundefined{tablewidth}{\newlength\tablewidth}\makeatother',
                r'\setlength\tablewidth\linewidth',
                r'\addtolength\tablewidth{-6\tabcolsep}',
                r'\renewcommand{\arraystretch}{1.4}',
                r'\begin{tabular}{lll}',

                r'\cline{2-2}',
                r'\multicolumn{1}{l|}{A1} & ' + \
                    r'\multicolumn{1}{l}{B1} & ' + \
                    r'\multicolumn{1}{|l}{C1} \\',

                r'\cline{2-3}',
                r'\multicolumn{1}{l|}{A2} & ' + \
                    r'\multicolumn{1}{l}{B2} & ' + \
                    r'\multicolumn{1}{|l}{C2} \\',
                r'\cline{2-3}',

                r'\end{tabular}\\',
                r'\vspace{4pt}',
                r''
                ))

        self.assertMultiLineEqual(self.convert(html), latex)

    def test_alignment_classes(self):
        html = '\n'.join((
                r'without width:',
                r'<table class="no-page-break">',
                r' <tr><td class="left">left</td></tr>',
                r' <tr><td class="center">center</td></tr>',
                r' <tr><td class="right">right</td></tr>',
                r'</table>',

                r'with width:',
                r'<table class="no-page-break">',
                r' <colgroup>',
                r'  <col width="100%" />',
                r' </colgroup>',
                r' <tr><td class="left">left</td></tr>',
                r' <tr><td class="center">center</td></tr>',
                r' <tr><td class="right">right</td></tr>',
                r' </tr>',
                r'</table>',
                ))

        latex = '\n'.join((
                r'without width: '
                r'\makeatletter\@ifundefined{tablewidth}{\newlength\tablewidth}\makeatother',
                r'\setlength\tablewidth\linewidth',
                r'\addtolength\tablewidth{-2\tabcolsep}',
                r'\renewcommand{\arraystretch}{1.4}',
                r'\begin{tabular}{l}',
                r'\multicolumn{1}{l}{left} \\',
                r'\multicolumn{1}{c}{center} \\',
                r'\multicolumn{1}{r}{right} \\',
                r'\end{tabular}\\',
                r'\vspace{4pt}',

                r' with width: '
                r'\makeatletter\@ifundefined{tablewidth}{\newlength\tablewidth}\makeatother',
                r'\setlength\tablewidth\linewidth',
                r'\addtolength\tablewidth{-2\tabcolsep}',
                r'\renewcommand{\arraystretch}{1.4}',
                r'\begin{tabular}{p{1.0\tablewidth}}',
                r'\multicolumn{1}{p{1.0\tablewidth}}{\raggedright left} \\',
                r'\multicolumn{1}{p{1.0\tablewidth}}{' + \
                    r'\centering center} \\',
                r'\multicolumn{1}{p{1.0\tablewidth}}{\raggedleft right} \\',
                r'\end{tabular}\\',
                r'\vspace{4pt}',
                r''
                ))

        self.assertMultiLineEqual(self.convert(html, count=2), latex)

    def test_indent2_class(self):
        html = '\n'.join((
                r'<table>',
                r'  <tr>',
                r'    <td width="50%" class="indent2">foo</td>',
                r'    <td width="50%">bar</td>',
                r'  </tr>',
                r'</table>',
                ))

        latex = '\n'.join((
                r'\makeatletter\@ifundefined{tablewidth}{\newlength\tablewidth}\makeatother',
                r'\setlength\tablewidth\linewidth',
                r'\addtolength\tablewidth{-4\tabcolsep}',
                r'\renewcommand{\arraystretch}{1.4}',
                r'\begin{tabular}{p{0.5\tablewidth}p{0.5\tablewidth}}',

                r'\multicolumn{1}{p{0.5\tablewidth}}{' + \
                    r'\hangindent 0.2cm\hspace{0.2cm} foo} & ' + \
                    r'\multicolumn{1}{p{0.5\tablewidth}}{' + \
                    r'bar} \\',

                r'\end{tabular}\\',
                r'\vspace{4pt}',
                r''
                ))

        self.assertMultiLineEqual(self.convert(html), latex)

    def test_indent10_class(self):
        html = '\n'.join((
                r'<table>',
                r'  <tr>',
                r'    <td width="50%" class="indent10">foo</td>',
                r'    <td width="50%">bar</td>',
                r'  </tr>',
                r'</table>',
                ))

        latex = '\n'.join((
                r'\makeatletter\@ifundefined{tablewidth}{\newlength\tablewidth}\makeatother',
                r'\setlength\tablewidth\linewidth',
                r'\addtolength\tablewidth{-4\tabcolsep}',
                r'\renewcommand{\arraystretch}{1.4}',
                r'\begin{tabular}{p{0.5\tablewidth}p{0.5\tablewidth}}',

                r'\multicolumn{1}{p{0.5\tablewidth}}{' + \
                    r'\hangindent 1cm\hspace{1cm} foo} & ' + \
                    r'\multicolumn{1}{p{0.5\tablewidth}}{' + \
                    r'bar} \\',

                r'\end{tabular}\\',
                r'\vspace{4pt}',
                r''
                ))

        self.assertMultiLineEqual(self.convert(html), latex)

    def test_bold_class(self):
        html = '\n'.join((
                r'<table>',
                r'  <tr>',
                r'    <td width="50%" class="bold">foo</td>',
                r'    <td width="50%" class="bold indent10">bar</td>',
                r'  </tr>',
                r'</table>',
                ))

        latex = '\n'.join((
                r'\makeatletter\@ifundefined{tablewidth}{\newlength\tablewidth}\makeatother',
                r'\setlength\tablewidth\linewidth',
                r'\addtolength\tablewidth{-4\tabcolsep}',
                r'\renewcommand{\arraystretch}{1.4}',
                r'\begin{tabular}{p{0.5\tablewidth}p{0.5\tablewidth}}',

                r'\multicolumn{1}{p{0.5\tablewidth}}{' + \
                    r'\textbf{foo}} & ' + \
                    r'\multicolumn{1}{p{0.5\tablewidth}}{' + \
                    r'\hangindent 1cm\hspace{1cm} \textbf{bar}} \\',

                r'\end{tabular}\\',
                r'\vspace{4pt}',
                r''
                ))

        self.assertMultiLineEqual(self.convert(html), latex)

    def test_grey_class(self):
        html = '\n'.join((
                r'<table>',
                r'  <tr>',
                r'    <td width="50%" class="grey">foo</td>',
                r'    <td width="50%">bar</td>',
                r'  </tr>',
                r'</table>',
                ))

        latex = '\n'.join((
                r'\makeatletter\@ifundefined{tablewidth}{\newlength\tablewidth}\makeatother',
                r'\setlength\tablewidth\linewidth',
                r'\addtolength\tablewidth{-4\tabcolsep}',
                r'\renewcommand{\arraystretch}{1.4}',
                r'\begin{tabular}{p{0.5\tablewidth}p{0.5\tablewidth}}',

                r'\multicolumn{1}{p{0.5\tablewidth}}{' + \
                    r'\textcolor{gray}{foo}} & ' + \
                    r'\multicolumn{1}{p{0.5\tablewidth}}{' + \
                    r'bar} \\',

                r'\end{tabular}\\',
                r'\vspace{4pt}',
                r''
                ))

        self.assertMultiLineEqual(self.convert(html, use_packages=['xcolor']),
                                  latex)

    def test_footnotesize_class(self):
        html = '\n'.join((
                r'<table>',
                r'  <tr>',
                r'    <td width="50%" class="footnotesize">foo</td>',
                r'    <td width="50%">bar</td>',
                r'  </tr>',
                r'</table>',
                ))

        latex = '\n'.join((
                r'\makeatletter\@ifundefined{tablewidth}{\newlength\tablewidth}\makeatother',
                r'\setlength\tablewidth\linewidth',
                r'\addtolength\tablewidth{-4\tabcolsep}',
                r'\renewcommand{\arraystretch}{1.4}',
                r'\begin{tabular}{p{0.5\tablewidth}p{0.5\tablewidth}}',

                r'\multicolumn{1}{p{0.5\tablewidth}}{' + \
                    r'\footnotesize foo} & ' + \
                    r'\multicolumn{1}{p{0.5\tablewidth}}{' + \
                    r'bar} \\',

                r'\end{tabular}\\',
                r'\vspace{4pt}',
                r''
                ))

        self.assertMultiLineEqual(self.convert(html), latex)

    def test_scriptsize_class(self):
        html = '\n'.join((
                r'<table>',
                r'  <tr>',
                r'    <td width="50%" class="scriptsize">foo</td>',
                r'    <td width="50%">bar</td>',
                r'  </tr>',
                r'</table>',
                ))

        latex = '\n'.join((
                r'\makeatletter\@ifundefined{tablewidth}{\newlength\tablewidth}\makeatother',
                r'\setlength\tablewidth\linewidth',
                r'\addtolength\tablewidth{-4\tabcolsep}',
                r'\renewcommand{\arraystretch}{1.4}',
                r'\begin{tabular}{p{0.5\tablewidth}p{0.5\tablewidth}}',

                r'\multicolumn{1}{p{0.5\tablewidth}}{' + \
                    r'\scriptsize foo} & ' + \
                    r'\multicolumn{1}{p{0.5\tablewidth}}{' + \
                    r'bar} \\',

                r'\end{tabular}\\',
                r'\vspace{4pt}',
                r''
                ))

        self.assertMultiLineEqual(self.convert(html), latex)

    def test_ragged_and_style_order(self):
        # \raggedright should be first, the \textbf
        html = '\n'.join((
                r'<table>',
                r'  <tr>',
                r'    <td width="50%" class="bold right">foo</td>',
                r'    <td width="50%">bar</td>',
                r'  </tr>',
                r'</table>',
                ))

        latex = '\n'.join((
                r'\makeatletter\@ifundefined{tablewidth}{\newlength\tablewidth}\makeatother',
                r'\setlength\tablewidth\linewidth',
                r'\addtolength\tablewidth{-4\tabcolsep}',
                r'\renewcommand{\arraystretch}{1.4}',
                r'\begin{tabular}{p{0.5\tablewidth}p{0.5\tablewidth}}',

                r'\multicolumn{1}{p{0.5\tablewidth}}{' + \
                    r'\raggedleft \textbf{foo}} & ' + \
                    r'\multicolumn{1}{p{0.5\tablewidth}}{' + \
                    r'bar} \\',

                r'\end{tabular}\\',
                r'\vspace{4pt}',
                r''
                ))

        self.assertMultiLineEqual(self.convert(html), latex)

    def test_more_cols_than_cells(self):
        # Regression test for when there are more <col> tags than <td> tags per row.
        html = '\n'.join((
                '<table class="plain" style="width: 100%;">',
                '  <colgroup>',
                '    <col width="17%" />',
                '    <col width="41%" />',
                '    <col width="35%" />',
                '  </colgroup>',
                '  <tbody>',
                '    <tr>',
                '      <th style="width: 30%;">Foo</th>',
                '      <th style="width: 70%;">Bar</th>',
                '    </tr>',
                '  </tbody>',
                '</table>'))

        latex = '\n'.join((
                r'\makeatletter\@ifundefined{tablewidth}{\newlength\tablewidth}\makeatother',
                r'\setlength\tablewidth\linewidth',
                r'\addtolength\tablewidth{-4\tabcolsep}',
                r'\renewcommand{\arraystretch}{1.4}',
                r'\begin{tabular}{p{0.17\tablewidth}p{0.41\tablewidth}}',
                r'\multicolumn{1}{p{0.3\tablewidth}}{\textbf{Foo}} & \multicolumn{1}{p{0.7\tablewidth}}{\textbf{Bar}} \\',
                r'\end{tabular}\\',
                r'\vspace{4pt}',
                r''
                ))

        self.maxDiff = None
        self.assertMultiLineEqual(self.convert(html), latex)

    def test_vertical_th_isnt_repeated(self):
        html = '\n'.join((
                r'<table class="plain" style="width: 100%;">',
                r'  <colgroup>',
                r'    <col width="17%" />',
                r'    <col width="41%" />',
                r'    <col width="35%" />',
                r'  </colgroup>',
                r'  <thead>',
                r'    <tr>',
                r'      <td style="width: 30%;">Foo</th>',
                r'      <td style="width: 70%;">Bar</th>',
                r'    </tr>',
                r'   </thead>',
                r'   <tbody>',
                20 * r'    <tr><th style="width: 30%;">Foo</th><td style="width: 70%;">Batz</td></tr>',
                r'  </tbody>',
                r'</table>'))

        latex_list = [
            r'\makeatletter\@ifundefined{tablewidth}{\newlength\tablewidth}\makeatother',
            r'\setlength\tablewidth\linewidth',
            r'\addtolength\tablewidth{-4\tabcolsep}',
            r'\renewcommand{\arraystretch}{1.4}',
            r'\begin{longtable}{p{0.17\tablewidth}p{0.41\tablewidth}}',
            r'\multicolumn{1}{p{0.3\tablewidth}}{\textbf{Foo}} & \multicolumn{1}{p{0.7\tablewidth}}{\textbf{Bar}} \\',
            r'\endhead',
        ]
        for counter in range(20):
            latex_list.append(r'\multicolumn{1}{p{0.3\tablewidth}}{\textbf{Foo}} & \multicolumn{1}{p{0.7\tablewidth}}{Batz} \\')
        latex_list.append(r'\end{longtable}\\')
        latex_list.append(r'\vspace{4pt}')
        latex_list.append(r'')
        latex = '\n'.join(latex_list)

        self.maxDiff = None
        self.assertMultiLineEqual(self.convert(html, use_packages=['longtable']), latex)


class TestLatexWidth(TestCase):

    def test_failing_convertion(self):
        with self.assertRaises(ValueError) as cm:
            table.LatexWidth.convert('123xy')

        self.assertEqual(
            str(cm.exception),
            'Could not convert string "123xy" to valid LatexWidth')

    def test_absolute_unit_width(self):
        width1 = table.LatexWidth.convert('100cm')
        self.assertEqual(width1.get(), r'100.0cm')
        self.assertEqual(str(width1), r'100.0cm')

        width2 = table.LatexWidth.convert('50.3cm')
        self.assertEqual(width2.get(), r'50.3cm')

        self.assertEqual((width1 + width2).get(),
                         r'150.3cm')

    def test_relative_width(self):
        width1 = table.LatexWidth.convert('10%')
        self.assertEqual(str(width1), r'0.1\tablewidth')

        width2 = table.LatexWidth.convert('25%')
        self.assertEqual(str(width2), r'0.25\tablewidth')

        self.assertEqual(str(width1 + width2),
                         r'0.35\tablewidth')

    def test_defaults_to_em(self):
        width1 = table.LatexWidth.convert('50')
        self.assertEqual(str(width1), r'50.0em')

        width2 = table.LatexWidth.convert('20.0em')
        self.assertEqual(str(width1 + width2),
                         r'70.0em')

    def test_cm_is_converted_to_mm_when_summing(self):
        width1 = table.LatexWidth.convert('10.3cm')
        width2 = table.LatexWidth.convert('20.1mm')
        self.assertEqual(str(width1 + width2),
                         r'123.1mm')
        self.assertEqual(str(width2 + width1),
                         r'123.1mm')

    def test_px_is_the_same_as_no_measure_unit(self):
        width_px = table.LatexWidth.convert('50px')
        width_no = table.LatexWidth.convert('50')

        self.assertEqual(str(width_no), str(width_px))

    def test_fails_if_summing_with_non_width(self):
        width = table.LatexWidth.convert('10.3cm')

        with self.assertRaises(ValueError) as cm:
            width = width + 5

        self.assertEqual(
            str(cm.exception),
            'Cannot accumulate LatexWidth with int')

    def test_fails_if_unmatching_units(self):
        width1 = table.LatexWidth.convert('10em')
        width2 = table.LatexWidth.convert('20mm')

        with self.assertRaises(ValueError) as cm:
            width1 = width1 + width2

        self.assertEqual(
            str(cm.exception),
            'Cannot accumulate LatexWidths with different units (em, mm)')

    def test_fails_if_unmatching_types(self):
        width1 = table.LatexWidth.convert('10em')
        width2 = table.LatexWidth.convert('20%')

        with self.assertRaises(ValueError) as cm:
            width1 = width1 + width2

        self.assertEqual(
            str(cm.exception),
            'Cannot accumulate relative and absolute widths.')


class TestParseHTMLStyleAttribute(TestCase):

    def test_empty(self):
        self.assertEqual(table.parse_html_style_attribute(''), {})

    def test_none(self):
        self.assertEqual(table.parse_html_style_attribute(None), {})

    def test_single_argument(self):
        self.assertEqual(table.parse_html_style_attribute('width: 10'),
                         {'width': '10'})

        self.assertEqual(table.parse_html_style_attribute('width:10;'),
                         {'width': '10'})

        self.assertEqual(table.parse_html_style_attribute('width : 10 ;'),
                         {'width': '10'})

        self.assertEqual(table.parse_html_style_attribute('width : 10 ;;'),
                         {'width': '10'})

    def test_multiple_arguments(self):
        self.assertEqual(table.parse_html_style_attribute(
                'width:10;height:5'),
                         {'width': '10',
                          'height': '5'})

        self.assertEqual(table.parse_html_style_attribute(
                'width : 10 ; height : 5 ;'),
                         {'width': '10',
                          'height': '5'})



from ftw.pdfgenerator.html2latex import subconverter
import re


class FootnoteConverter(subconverter.SubConverter):

    pattern = (r'<span [^>]*?class="(?P<class>[^"]*?footnote[^"]*?)"[^>]*?>'
               '(?P<content>.*?)</span>')

    def __call__(self):
        if 'footnote' not in self.match.group('class').split():
            return

        fntext_match = re.search(r'data-footnote="(?P<footnoteText>[^"]*)"', self.get_html())
        if not fntext_match:
            return
        footnoteText = fntext_match.group('footnoteText')

        content = self.match.group('content')

        content = self.converter.convert(content)
        footnoteText = self.converter.convert(footnoteText)
        latex = r'%s\footnote{%s}' % (content, footnoteText)

        self.replace_and_lock(latex)



from ftw.pdfgenerator import _
from zope.component.hooks import getSite
from zope.i18n import translate


ENVIRONMENTS = {
    'table': {
        'label': _(u'enviroment_table', default=u'Table'),
        'index': 'lot',
        'counter': 'table'},

    'figure': {
        'label': _(u'enviroment_figure', default=u'Figure'),
        'index': 'lof',
        'counter': 'figure'}}


def generate_manual_caption(caption, environment, show_in_index=True):
    """In LaTeX, captions are used within floating environmonts (such as
    "table" or "figure").
    Since we generate LaTeX automatically we are usually not using floating
    environments so that LaTeX does not reposition images and tables.
    Because of that we cannot use the standard \caption command, we need
    to do it manually.

    This helper function generates the LaTeX code for placing a caption
    outside a floating environment.

    Arguments:
    caption -- The caption text as LaTeX.
    enviroment -- The type of environment (either "table" or "figure").
    """

    if not caption:
        return ''

    if environment not in ENVIRONMENTS.keys():
        raise ValueError('Environment must be one of "%s". Got "%s".' % (
                '", "'.join(ENVIRONMENTS.keys()), environment))

    env = ENVIRONMENTS[environment]
    label = translate(env['label'], context=getSite().REQUEST).encode('utf-8')

    latex = [r'\begin{center}']

    if show_in_index:
        latex.extend([
                r'\addtocounter{%s}{1}' % env['counter'],

                r'\addcontentsline{%s}{%s}{' % (
                    env['index'], env['counter']) +
                r'\protect\numberline ' +
                r'{\thechapter.\arabic{%s}}' % env['counter'] +
                r'{\ignorespaces %s}' % caption +
                r'}',

                r'%s \thechapter.\arabic{%s}: %s' % (
                    label, env['counter'], caption),
                ])

    else:
        latex.append(caption)

    latex.extend([
            r'\end{center}',
            r'',
            ])

    return '\n'.join(latex)



from ftw.pdfgenerator.browser.views import DebugPDFGeneratorView
from ftw.pdfgenerator.interfaces import DEBUG_MODE_SESSION_KEY
from ftw.pdfgenerator.testing import PDFGENERATOR_ZCML_LAYER
from plone.mocktestcase.dummy import Dummy
from unittest2 import TestCase
from zope.component import getMultiAdapter
from zope.interface import directlyProvides
from zope.publisher.interfaces.browser import IDefaultBrowserLayer


class TestDebugPDFGeneratorView(TestCase):

    layer = PDFGENERATOR_ZCML_LAYER

    def test_component_registered(self):
        request = Dummy()
        directlyProvides(request, IDefaultBrowserLayer)
        view = getMultiAdapter((object(), request), name='debug-pdf')
        self.assertTrue(isinstance(view, DebugPDFGeneratorView))

    def test_toggle(self):
        session = {}
        request = Dummy(SESSION=session)
        directlyProvides(request, IDefaultBrowserLayer)

        view = getMultiAdapter((object(), request), name='debug-pdf')

        self.assertEquals(session.get(DEBUG_MODE_SESSION_KEY, False), False)

        self.assertEquals(view(), 'PDFGenerator debug mode is now enabled')
        self.assertEquals(session.get(DEBUG_MODE_SESSION_KEY, False), True)

        self.assertEquals(view(), 'PDFGenerator debug mode is now disabled')
        self.assertEquals(session.get(DEBUG_MODE_SESSION_KEY, False), False)



from Products.Five import BrowserView
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile
from ftw.pdfgenerator.interfaces import DEBUG_MODE_SESSION_KEY
from ftw.pdfgenerator.interfaces import IPDFAssembler
from zope.component import getMultiAdapter


class ExportPDFView(BrowserView):
    """Export a PDF with default settings. If the user is a Admin (if he
    has Manage portal permission), a additional form will be shown, where
    he can selected the desired output format (PDF, LaTeX only or ZIP).
    """

    index = ViewPageTemplateFile('export_pdf.pt')

    def __call__(self):
        if self.allow_alternate_output():
            if self.request.get('submitted', False):
                output = self.request.get('output')
                return self.export(output)

            return self.index()

        return self.export()

    def allow_alternate_output(self):
        """For selecting the output format, the user must have Manage portal
        permission.
        """
        user = self.context.portal_membership.getAuthenticatedMember()
        if user.has_permission('cmf.ManagePortal', self.context):
            return True

        elif self.request.SESSION.get(DEBUG_MODE_SESSION_KEY, False):
            return True

        else:
            return False

    def export(self, output='pdf'):
        assembler = getMultiAdapter((self.context, self.request),
                                    IPDFAssembler)

        if output == 'pdf':
            return assembler.build_pdf(**self.get_build_arguments())

        elif output == 'latex':
            return assembler.build_latex(**self.get_build_arguments())

        elif output == 'zip':
            return assembler.build_zip(**self.get_build_arguments())

        else:
            raise ValueError('Unkown output "%s"' % output)

    def get_build_arguments(self):
        return {'request': self.request}


class DebugPDFGeneratorView(BrowserView):
    """This view allows to toggle the pdfgenerator debug mode.
    When the debug mode is enabled, the user is able to select the
    output.
    """

    def __call__(self):
        session = self.request.SESSION

        session[DEBUG_MODE_SESSION_KEY] = not session.get(
            DEBUG_MODE_SESSION_KEY, False)

        return 'PDFGenerator debug mode is now %s' % (
            session.get(DEBUG_MODE_SESSION_KEY) and 'enabled' or 'disabled')



from ftw.pdfgenerator.interfaces import ITemplating
from ftw.pdfgenerator.utils import baseclasses
from mako.lookup import TemplateLookup
from zope.interface import implements
import inspect
import os
import posixpath
import re


class BaseTemplating(object):
    """For details see `ftw.pdfgenerator.interfaces.ITemplating`
    """
    implements(ITemplating)

    template_directories = []

    def get_template_directories(self):
        dirs = []

        if self.__class__.__module__ == 'Products.Five.metaclass':
            # self is a browser view which is based on a metaclass.
            # We need to get rid of the metaclass so that we know
            # where the class is defined.
            classes = baseclasses(self.__class__)[1:]
        else:
            classes = baseclasses(self.__class__)

        for cls in classes:
            local_paths = getattr(cls, 'template_directories', None)

            if local_paths is None:
                continue

            elif not hasattr(local_paths, '__iter__'):
                classname = cls.__name__
                raise ValueError(classname + '.template_directories '
                                 'should be a list or None.')

            for path in reversed(local_paths):
                if not os.path.isabs(path):
                    path = os.path.join(
                        os.path.dirname(inspect.getsourcefile(cls)), path)

                # Make an acces for checking if it exists. This will raise
                # an OSError if the path does not exist.
                os.listdir(path)

                if path not in dirs:
                    dirs.append(path)

        return dirs

    def get_template(self, filename):
        for directory in self.get_template_directories():
            path = os.path.join(directory, filename)
            if os.path.exists(path):
                return open(path, 'rb').read()

        return None

    def render_template(self, filename, **kwargs):
        # render_template should be implemented by a subclass, which will
        # be more specific about the template rendering engine.
        raise NotImplementedError(
            'render_template() is not implemented on BaseTemplating.')


class MakoTemplating(BaseTemplating):
    """Provides a mako templating integration.
    """

    def __init__(self):
        super(MakoTemplating, self).__init__()
        self._mako_template_lookup = None

    @property
    def template_lookup(self):
        """Returns a mako TemplateLookup object, which has all current
        template directories configured.
        """

        if getattr(self, '_mako_template_lookup', None) is None:
            dirs = self.get_template_directories()
            self._mako_template_lookup = TemplateLookup(
                directories=dirs,
                default_filters=['decode.utf8'],
                input_encoding='utf-8')

        return self._mako_template_lookup

    def render_template(self, filename, **kwargs):
        """Renders a mako template. Additional rendering arguments may be
        passed as keyword arguments as if the mako template was called
        directly.
        """

        template = self.template_lookup.get_template(filename)
        kwargs['view'] = self
        return template.render(**kwargs)

    def get_raw_template(self, name):
        """Returns the contents of a template file without parsing it.
        """
        uri = re.sub(r'^\/+', '', name)
        for dir_ in self.template_lookup.directories:
            srcfile = posixpath.normpath(posixpath.join(dir_, uri))
            if os.path.isfile(srcfile):
                return open(srcfile).read()
        return None
